# Import Necessary Libraries
import tensorflow as tf

def Conv_Block(inputs, model_width, kernel, multiplier=1, bottleneck=False):
    # 1D Convolutional Block
    if bottleneck:
        # Bottleneck design: use 1x1 conv to reduce channels before 3x3 conv
        x = tf.keras.layers.Conv1D(model_width * multiplier // 4, 1, padding='same', kernel_initializer="he_normal")(inputs)
        x = tf.keras.layers.BatchNormalization()(x)
        x = tf.keras.layers.Activation('relu')(x)
        x = tf.keras.layers.Conv1D(model_width * multiplier, kernel, padding='same', kernel_initializer="he_normal")(x)
    else:
        # Standard convolution
        x = tf.keras.layers.Conv1D(model_width * multiplier, kernel, padding='same', kernel_initializer="he_normal")(inputs)
    
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.layers.Activation('relu')(x)

    return x


def Concat_Block(input1, *argv):
    # Concatenation Block from the Keras Library
    cat = input1
    for arg in range(0, len(argv)):
        cat = tf.keras.layers.concatenate([cat, argv[arg]], axis=-1)

    return cat


def upConv_Block(inputs):
    # 1D UpSampling Block
    up = tf.keras.layers.UpSampling1D(size=2)(inputs)

    return up


def trans_conv1D(inputs, model_width, multiplier, strides):
    # 1D Transposed Convolutional Block, used instead of UpSampling
    x = tf.keras.layers.Conv1DTranspose(model_width * multiplier, 4, strides=strides, padding='same')(inputs)  # Stride = 2, Kernel Size = 4
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.layers.Activation('relu')(x)

    return x


def Feature_Extraction_Block(inputs, model_width, feature_number):
    # Feature Extraction Block for the AutoEncoder Mode
    shape = inputs.shape
    latent = tf.keras.layers.Flatten()(inputs)
    latent = tf.keras.layers.Dense(feature_number, name='features')(latent)
    latent = tf.keras.layers.Dense(model_width * shape[1])(latent)
    latent = tf.keras.layers.Reshape((shape[1], model_width))(latent)

    return latent


def Attention_Block(skip_connection, gating_signal, num_filters, multiplier):
    # Attention Block
    conv1x1_1 = tf.keras.layers.Conv1D(num_filters * multiplier, 1, strides=2)(skip_connection)
    conv1x1_1 = tf.keras.layers.BatchNormalization()(conv1x1_1)
    conv1x1_2 = tf.keras.layers.Conv1D(num_filters * multiplier, 1, strides=1)(gating_signal)
    conv1x1_2 = tf.keras.layers.BatchNormalization()(conv1x1_2)
    conv1_2 = tf.keras.layers.add([conv1x1_1, conv1x1_2])
    conv1_2 = tf.keras.layers.Activation('relu')(conv1_2)
    conv1_2 = tf.keras.layers.Conv1D(1, 1, strides=1)(conv1_2)
    conv1_2 = tf.keras.layers.BatchNormalization()(conv1_2)
    conv1_2 = tf.keras.layers.Activation('sigmoid')(conv1_2)
    resampler = upConv_Block(conv1_2)
    out = skip_connection * resampler

    return out


class TernausNet:
    def __init__(self, length, num_channel, model_width, ds=0, ae=0, ag=0, problem_type='Regression',
                 output_nums=1, feature_number=1024, is_transconv=True):
        # length: Input Signal Length
        # model_width: Width of the Input Layer of the Model
        # num_channel: Number of Channels allowed by the Model
        # problem_type: Classification (Binary or Multiclass) or Regression
        # output_nums: Output Classes (Classification Mode) or Features (Regression Mode)
        # ds: Checks where Deep Supervision is active or not, either 0 or 1 [Default value set as 0]
        # ae: Enables or diables the AutoEncoder Mode, either 0 or 1 [Default value set as 0]
        # ag: Checks where Attention Guided is active or not, either 0 or 1 [Default value set as 0]
        # feature_number: Number of Features or Embeddings to be extracted from the AutoEncoder in the A_E Mode
        # is_transconv: (TRUE - Transposed Convolution, FALSE - UpSampling) in the Encoder Layer
        self.length = length
        self.num_channel = num_channel
        self.model_width = model_width
        self.problem_type = problem_type
        self.output_nums = output_nums
        self.D_S = ds
        self.A_E = ae
        self.A_G = ag
        self.feature_number = feature_number
        self.is_transconv = is_transconv

    def TernausNet11(self):
        """TernausNet11 Model Design - UNet with VGG11 Backend"""
        if self.length == 0 or self.model_width == 0 or self.num_channel == 0:
            raise ValueError("Please Check the Values of the Input Parameters!")

        convs = {}
        levels = []

        # Encoding
        inputs = tf.keras.Input((self.length, self.num_channel))
        conv = Conv_Block(inputs, self.model_width, 3, 2 ** 0)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv0"] = conv
        conv = Conv_Block(pool, self.model_width, 3, 2 ** 1)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv1"] = conv

        for i in range(2, 4):
            conv = Conv_Block(pool, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** i)
            pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
            convs["conv%s" % i] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv4"] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Collect Latent Features or Embeddings from AutoEncoders
        if self.A_E == 1:
            latent = Feature_Extraction_Block(pool, self.model_width, self.feature_number)
            conv = Conv_Block(latent, self.model_width, 3, 2 ** 3)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Decoding
        convs_list = list(convs.values())
        deconv = conv
        for j in range(0, 5):
            skip_connection = convs_list[4 - j]
            if self.A_G == 1:
                skip_connection = Attention_Block(convs_list[4 - j], deconv, self.model_width, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            if self.is_transconv:
                deconv = Concat_Block(trans_conv1D(deconv, self.model_width, 2 ** (4 - j), 2), skip_connection)
            elif not self.is_transconv:
                deconv = Concat_Block(upConv_Block(deconv), skip_connection)
            if self.D_S == 1:
                level = tf.keras.layers.Conv1D(1, 1, name=f'level{4 - j}')(deconv)
                levels.append(level)

        deconv = Conv_Block(deconv, self.model_width, 3, 2 ** 0)

        # Output
        outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='linear', name="out")(deconv)
        if self.problem_type == 'Classification':
            outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='softmax', name="out")(deconv)

        model = tf.keras.Model(inputs=[inputs], outputs=[outputs])

        if self.D_S == 1:
            levels.append(outputs)
            levels.reverse()
            model = tf.keras.Model(inputs=[inputs], outputs=levels)

        return model

    def TernausNet13(self):
        """TernausNet13 Model Design - UNet with VGG13 Backend"""
        if self.length == 0 or self.model_width == 0 or self.num_channel == 0:
            raise ValueError("Please Check the Values of the Input Parameters!")

        convs = {}
        levels = []

        # Encoding
        inputs = tf.keras.Input((self.length, self.num_channel))
        conv = Conv_Block(inputs, self.model_width, 3, 2 ** 0)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 0)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv0"] = conv
        conv = Conv_Block(pool, self.model_width, 3, 2 ** 1)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 1)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv1"] = conv

        for i in range(2, 4):
            conv = Conv_Block(pool, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** i)
            pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
            convs["conv%s" % i] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv4"] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Collect Latent Features or Embeddings from AutoEncoders
        if self.A_E == 1:
            latent = Feature_Extraction_Block(pool, self.model_width, self.feature_number)
            conv = Conv_Block(latent, self.model_width, 3, 2 ** 3)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Decoding
        convs_list = list(convs.values())
        deconv = conv
        for j in range(0, 5):
            skip_connection = convs_list[4 - j]
            if self.A_G == 1:
                skip_connection = Attention_Block(convs_list[4 - j], deconv, self.model_width, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            if self.is_transconv:
                deconv = Concat_Block(trans_conv1D(deconv, self.model_width, 2 ** (4 - j), 2), skip_connection)
            elif not self.is_transconv:
                deconv = Concat_Block(upConv_Block(deconv), skip_connection)
            if self.D_S == 1:
                level = tf.keras.layers.Conv1D(1, 1, name=f'level{4 - j}')(deconv)
                levels.append(level)

        deconv = Conv_Block(deconv, self.model_width, 3, 2 ** 0)

        # Output
        outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='linear', name="out")(deconv)
        if self.problem_type == 'Classification':
            outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='softmax', name="out")(deconv)

        model = tf.keras.Model(inputs=[inputs], outputs=[outputs])

        if self.D_S == 1:
            levels.append(outputs)
            levels.reverse()
            model = tf.keras.Model(inputs=[inputs], outputs=levels)

        return model

    def TernausNet16(self):
        """TernausNet16 Model Design - UNet with VGG16 Backend"""
        if self.length == 0 or self.model_width == 0 or self.num_channel == 0:
            raise ValueError("Please Check the Values of the Input Parameters!")

        convs = {}
        levels = []

        # Encoding
        inputs = tf.keras.Input((self.length, self.num_channel))
        conv = Conv_Block(inputs, self.model_width, 3, 2 ** 0)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 0)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv0"] = conv
        conv = Conv_Block(pool, self.model_width, 3, 2 ** 1)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 1)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv1"] = conv

        for i in range(2, 4):
            conv = Conv_Block(pool, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 1, 2 ** i)
            pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
            convs["conv%s" % i] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 1, 2 ** 3)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv4"] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Collect Latent Features or Embeddings from AutoEncoders
        if self.A_E == 1:
            latent = Feature_Extraction_Block(pool, self.model_width, self.feature_number)
            conv = Conv_Block(latent, self.model_width, 3, 2 ** 3)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Decoding
        convs_list = list(convs.values())

        deconv = conv
        for j in range(0, 5):
            skip_connection = convs_list[4 - j]
            if self.A_G == 1:
                skip_connection = Attention_Block(convs_list[4 - j], deconv, self.model_width, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            if self.is_transconv:
                deconv = Concat_Block(trans_conv1D(deconv, self.model_width, 2 ** (4 - j), 2), skip_connection)
            elif not self.is_transconv:
                deconv = Concat_Block(upConv_Block(deconv), skip_connection)
            if self.D_S == 1:
                level = tf.keras.layers.Conv1D(1, 1, name=f'level{4 - j}')(deconv)
                levels.append(level)

        deconv = Conv_Block(deconv, self.model_width, 3, 2 ** 0)

        # Output
        outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='linear', name="out")(deconv)
        if self.problem_type == 'Classification':
            outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='softmax', name="out")(deconv)

        model = tf.keras.Model(inputs=[inputs], outputs=[outputs])

        if self.D_S == 1:
            levels.append(outputs)
            levels.reverse()
            model = tf.keras.Model(inputs=[inputs], outputs=levels)

        return model

    def TernausNet19(self):
        """TernausNet19 Model Design - UNet with VGG19 Backend"""
        if self.length == 0 or self.model_width == 0 or self.num_channel == 0:
            raise ValueError("Please Check the Values of the Input Parameters!")

        convs = {}
        levels = []

        # Encoding
        inputs = tf.keras.Input((self.length, self.num_channel))
        conv = Conv_Block(inputs, self.model_width, 3, 2 ** 0)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 0)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv0"] = conv
        conv = Conv_Block(pool, self.model_width, 3, 2 ** 1)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 1)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv1"] = conv

        for i in range(2, 4):
            conv = Conv_Block(pool, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** i)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** i)
            pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
            convs["conv%s" % i] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)
        pool = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding="valid")(conv)
        convs["conv4"] = conv

        conv = Conv_Block(pool, self.model_width, 3, 2 ** 3)
        conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Collect Latent Features or Embeddings from AutoEncoders
        if self.A_E == 1:
            latent = Feature_Extraction_Block(pool, self.model_width, self.feature_number)
            conv = Conv_Block(latent, self.model_width, 3, 2 ** 3)
            conv = Conv_Block(conv, self.model_width, 3, 2 ** 3)

        # Decoding
        convs_list = list(convs.values())

        deconv = conv
        for j in range(0, 5):
            skip_connection = convs_list[4 - j]
            if self.A_G == 1:
                skip_connection = Attention_Block(convs_list[4 - j], deconv, self.model_width, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            deconv = Conv_Block(deconv, self.model_width, 3, 2 ** (4 - j))
            if self.is_transconv:
                deconv = Concat_Block(trans_conv1D(deconv, self.model_width, 2 ** (4 - j), 2), skip_connection)
            elif not self.is_transconv:
                deconv = Concat_Block(upConv_Block(deconv), skip_connection)
            if self.D_S == 1:
                level = tf.keras.layers.Conv1D(1, 1, name=f'level{4 - j}')(deconv)
                levels.append(level)

        deconv = Conv_Block(deconv, self.model_width, 3, 2 ** 0)

        # Output
        outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='linear', name="out")(deconv)
        if self.problem_type == 'Classification':
            outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='softmax', name="out")(deconv)

        model = tf.keras.Model(inputs=[inputs], outputs=[outputs])

        if self.D_S == 1:
            levels.append(outputs)
            levels.reverse()
            model = tf.keras.Model(inputs=[inputs], outputs=levels)

        return model


if __name__ == '__main__':
    # Configurations
    num_channel = 1
    length = 1024
    model_width = 64
    D_S = 1
    A_E = 1
    A_G = 1
    feature_number = 1024
    model_name = 'TernausNet16'
    # Build model for TernausNet
    Model = TernausNet(length, num_channel, model_width, problem_type='Regression', output_nums=1, ds=D_S, ae=A_E, ag=A_G, is_transconv=True).TernausNet16()
    Model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=0.0003), loss=tf.keras.losses.MeanAbsoluteError(), metrics=tf.keras.metrics.MeanSquaredError())
    Model.summary()



# import math
# from typing import List, Tuple, Dict

# import torch
# import torch.nn as nn
# import torch.nn.functional as F

# # -----------------------------------------------------------------------------
# #  Light‑weight Keras → PyTorch helpers
# # -----------------------------------------------------------------------------

# class DepthwiseConv1d(nn.Conv1d):
#     """Depth‑wise 1‑D convolution (grouped conv where groups=in_channels)."""

#     def __init__(self, in_channels: int, kernel_size: int, depth_multiplier: int = 1, **conv_kwargs):
#         super().__init__(
#             in_channels,
#             in_channels * depth_multiplier,
#             kernel_size,
#             groups=in_channels,
#             **conv_kwargs,
#         )


# class SeparableConv1d(nn.Module):
#     """Separable 1‑D convolution = depth‑wise + point‑wise."""

#     def __init__(self, in_channels: int, out_channels: int, kernel_size: int, depth_multiplier: int = 1, **conv_kwargs):
#         super().__init__()
#         self.depthwise = DepthwiseConv1d(in_channels, kernel_size, depth_multiplier, **conv_kwargs)
#         self.pointwise = nn.Conv1d(in_channels * depth_multiplier, out_channels, kernel_size=1)

#     def forward(self, x):  # (N,C,L)
#         return self.pointwise(self.depthwise(x))


# class SpatialDropout1d(nn.Dropout2d):
#     """TensorFlow‑like spatial dropout along the channel axis."""

#     def forward(self, x):  # (N,C,L)
#         x = x.unsqueeze(3)  # (N,C,L,1)
#         x = super().forward(x)
#         return x.squeeze(3)


# class GlobalAveragePooling1d(nn.Module):
#     def forward(self, x):  # (N,C,L)
#         return x.mean(dim=2)


# # -----------------------------------------------------------------------------
# #  Generic building blocks used by all architectures
# # -----------------------------------------------------------------------------

# def _activation(name: str):
#     if name == "swish":
#         return nn.SiLU()
#     elif name == "selu":
#         return nn.SELU()
#     elif name == "elu":
#         return nn.ELU()
#     else:
#         return nn.ReLU() if name == "relu" else getattr(nn, name.capitalize())()


# def _norm(method: str, num_features: int):
#     if method == "LayerNormalization":
#         return nn.LayerNorm(num_features)
#     return nn.BatchNorm1d(num_features)


# # -----------------------------------------------------------------------------
# #  Primitive model builders (DECON, NICON, Transformer)
# # -----------------------------------------------------------------------------


# def _build_decon(input_shape: Tuple[int, int], params: Dict, *, num_classes: int = 1):
#     """Port of the TensorFlow `decon` family to PyTorch.

#     This covers both regression (num_classes==1) and classification
#     (num_classes>=2, including binary with sigmoid).
#     """
#     c, _ = input_shape  # channels, length
#     layers: List[nn.Module] = [SpatialDropout1d(params.get("spatial_dropout", 0.2))]

#     def _dconv(kernel, depth_multiplier):
#         return DepthwiseConv1d(c, kernel, depth_multiplier=depth_multiplier, padding="same")

#     # First block (7‑7)
#     layers += [
#         _dconv(7, 2), _activation("relu"),
#         _dconv(7, 2), _activation("relu"),
#         nn.MaxPool1d(2, 2), nn.BatchNorm1d(c * 2),
#     ]

#     # Second block (5‑5)
#     layers += [
#         _dconv(5, 2), _activation("relu"),
#         _dconv(5, 2), _activation("relu"),
#         nn.MaxPool1d(2, 2), nn.BatchNorm1d(c * 4),
#     ]

#     # Third block (9‑9)
#     layers += [
#         _dconv(9, 2), _activation("relu"),
#         _dconv(9, 2), _activation("relu"),
#         nn.MaxPool1d(2, 2), nn.BatchNorm1d(c * 8),
#     ]

#     # Separable + ordinary conv
#     layers += [
#         SeparableConv1d(c * 8, 64, 3, depth_multiplier=1, padding="same"), _activation("relu"),
#         nn.Conv1d(64, 32, 3, padding="same"), _activation("relu"),
#         nn.MaxPool1d(5, 3), SpatialDropout1d(0.1),
#         nn.Flatten(),
#         nn.Linear(32 * math.ceil(input_shape[1] / 8 / 3), 128), _activation("relu"),
#         nn.Linear(128, 32), _activation("relu"),
#         nn.Dropout(0.2),
#     ]

#     # Output
#     if num_classes == 1:
#         layers.append(nn.Linear(32, 1))
#     elif num_classes == 2:
#         layers.append(nn.Linear(32, 1))
#         layers.append(nn.Sigmoid())
#     else:
#         layers.append(nn.Linear(32, num_classes))
#         layers.append(nn.Softmax(dim=1))

#     return nn.Sequential(*layers)


# def _build_transformer(input_shape: Tuple[int, int], params: Dict, *, num_classes: int = 1):
#     """1‑D Transformer with `nn.TransformerEncoder` blocks."""

#     c, seq_len = input_shape
#     head_size = params.get("head_size", 16)
#     num_heads = params.get("num_heads", 2)
#     ff_dim = params.get("ff_dim", 8)
#     num_blocks = params.get("num_transformer_blocks", 1)
#     dropout = params.get("dropout", 0.05)

#     d_model = head_size * num_heads

#     # Project the input channels to d_model, keeping length dimension
#     proj_in = nn.Conv1d(c, d_model, 1)

#     # Transformer encoder expects (L, N, E)
#     encoder_layer = nn.TransformerEncoderLayer(
#         d_model=d_model,
#         nhead=num_heads,
#         dim_feedforward=ff_dim * d_model,
#         batch_first=False,
#         dropout=dropout,
#         activation="relu",
#     )
#     encoder = nn.TransformerEncoder(encoder_layer, num_layers=num_blocks)

#     pooling = GlobalAveragePooling1d()

#     mlp_units = params.get("mlp_units", [32, 8])
#     mlp_dropout = params.get("mlp_dropout", 0.1)
#     mlp: List[nn.Module] = []
#     in_size = d_model
#     for dim in mlp_units:
#         mlp += [nn.Linear(in_size, dim), _activation("relu"), nn.Dropout(mlp_dropout)]
#         in_size = dim

#     # Output layer
#     if num_classes == 1:
#         out_layer = nn.Linear(in_size, 1)
#     elif num_classes == 2:
#         out_layer = nn.Sequential(nn.Linear(in_size, 1), nn.Sigmoid())
#     else:
#         out_layer = nn.Sequential(nn.Linear(in_size, num_classes), nn.Softmax(dim=1))

#     class _Transformer(nn.Module):
#         def __init__(self):
#             super().__init__()
#             self.proj_in = proj_in
#             self.encoder = encoder
#             self.pool = pooling
#             self.mlp = nn.Sequential(*mlp)
#             self.out = out_layer

#         def forward(self, x):  # x: (N, C, L)
#             z = self.proj_in(x)  # (N, d_model, L)
#             z = z.permute(2, 0, 1)  # (L, N, d_model)
#             z = self.encoder(z)
#             z = z.permute(1, 2, 0)  # (N, d_model, L)
#             z = self.pool(z)  # (N, d_model)
#             z = self.mlp(z)
#             return self.out(z)

#     return _Transformer()


# # -----------------------------------------------------------------------------
# #  Public API — function names must match the original tensorflow file
# # -----------------------------------------------------------------------------

# from nirs4all.utils.backend_utils import framework

# # --- DECON family ------------------------------------------------------------

# @framework("pytorch")
# def decon(input_shape, params={}):
#     return _build_decon(input_shape, params, num_classes=1)


# @framework("pytorch")
# def decon_classification(input_shape, num_classes=2, params={}):
#     return _build_decon(input_shape, params, num_classes=num_classes)


# @framework("pytorch")
# def decon_Sep(input_shape, params={}):
#     # Alias to standard decon for now – separable path is handled internally
#     return _build_decon(input_shape, params, num_classes=1)


# @framework("pytorch")
# def decon_Sep_classification(input_shape, num_classes=2, params={}):
#     return _build_decon(input_shape, params, num_classes=num_classes)


# # --- Transformer family ------------------------------------------------------

# @framework("pytorch")
# def transformer(input_shape, params={}):
#     return _build_transformer(input_shape, params, num_classes=1)


# @framework("pytorch")
# def transformer_VG(input_shape, params={}):
#     # Passthrough with different defaults handled in params already
#     return _build_transformer(input_shape, params, num_classes=1)


# @framework("pytorch")
# def transformer_classification(input_shape, num_classes=2, params={}):
#     return _build_transformer(input_shape, params, num_classes=num_classes)


# @framework("pytorch")
# def transformer_VG_classification(input_shape, num_classes=2, params={}):
#     return _build_transformer(input_shape, params, num_classes=num_classes)


# # --- NICON family (simplified) ----------------------------------------------

# def _build_nicon(input_shape, params, *, num_classes: int = 1):
#     c, _ = input_shape
#     layers: List[nn.Module] = [SpatialDropout1d(params.get("spatial_dropout", 0.08))]
#     layers += [
#         nn.Conv1d(c, params.get("filters1", 8), kernel_size=15, stride=5), _activation("selu"),
#         nn.Dropout(params.get("dropout_rate", 0.2)),
#         nn.Conv1d(params.get("filters1", 8), params.get("filters2", 64), kernel_size=21, stride=3), _activation("relu"),
#         nn.BatchNorm1d(params.get("filters2", 64)),
#         nn.Conv1d(params.get("filters2", 64), params.get("filters3", 32), kernel_size=5, stride=3), _activation("elu"),
#         nn.BatchNorm1d(params.get("filters3", 32)),
#         nn.Flatten(),
#         nn.Linear(params.get("filters3", 32), params.get("dense_units", 16)), _activation("sigmoid"),
#     ]

#     if num_classes == 1:
#         layers.append(nn.Linear(params.get("dense_units", 16), 1))
#     elif num_classes == 2:
#         layers += [nn.Linear(params.get("dense_units", 16), 1), nn.Sigmoid()]
#     else:
#         layers += [nn.Linear(params.get("dense_units", 16), num_classes), nn.Softmax(dim=1)]

#     return nn.Sequential(*layers)


# @framework("pytorch")
# def nicon(input_shape, params={}):
#     return _build_nicon(input_shape, params, num_classes=1)


# @framework("pytorch")
# def nicon_classification(input_shape, num_classes=2, params={}):
#     return _build_nicon(input_shape, params, num_classes=num_classes)


# @framework("pytorch")
# def nicon_VG(input_shape, params={}):
#     return _build_nicon(input_shape, params, num_classes=1)


# @framework("pytorch")
# def nicon_VG_classification(input_shape, num_classes=2, params={}):
#     return _build_nicon(input_shape, params, num_classes=num_classes)


# # --- Customizable variants ---------------------------------------------------

# @framework("pytorch")
# def customizable_nicon(input_shape, params={}):
#     return _build_nicon(input_shape, params, num_classes=1)


# @framework("pytorch")
# def customizable_nicon_classification(input_shape, num_classes=2, params={}):
#     return _build_nicon(input_shape, params, num_classes=num_classes)


# @framework("pytorch")
# def customizable_decon(input_shape, params={}):
#     return _build_decon(input_shape, params, num_classes=1)


# @framework("pytorch")
# def decon_layer_classification(input_shape, num_classes=2, params={}):
#     return _build_decon(input_shape, params, num_classes=num_classes)



from typing import Any, Dict, TYPE_CHECKING

from sklearn.base import TransformerMixin

from nirs4all.controllers.controller import OperatorController
from nirs4all.controllers.registry import register_controller

if TYPE_CHECKING:
    from nirs4all.pipeline.runner import PipelineRunner
    from nirs4all.spectra.spectra_dataset import SpectroDataset


@register_controller
class SampleAugmentationController(OperatorController):
    priority = 10

    @classmethod
    def matches(cls, step: Any, operator: Any, keyword: str) -> bool:
        return keyword == "sample_augmentation"

    @classmethod
    def use_multi_source(cls) -> bool:
        """Check if the operator supports multi-source datasets."""
        return True

    def execute(
        self,
        step: Any,
        operator: Any,
        dataset: 'SpectroDataset',
        context: Dict[str, Any],
        runner: 'PipelineRunner',
        source: int = -1
    ):
        print(f"Executing sample augmentation for step: {step}, keyword: {context.get('keyword', '')}, source: {source}")

        # Apply the transformer to the dataset
        try:
            contexts = []
            steps = []
            for i, operation in enumerate(step["sample_augmentation"]):
                augmentation_id=f"aug_{i}"
                if operation is None:
                    contexts.append(context)
                    steps.append(None)
                    continue
                local_context = context.copy()
                local_context["origin"] = None
                local_context["partition"] = "train"
                dataset.augment_samples(local_context, 1, augmentation_id=augmentation_id)
                local_context = context.copy()
                local_context["augmentation"] = augmentation_id
                contexts.append(local_context)
                steps.append(operation)

            runner.run_steps(steps, dataset, contexts, execution="sequential")
            return context

        except Exception as e:
            print(f"❌ Error applying transformation: {e}")
            raise


from ..augmentation.random import (Random_X_Operation,
                                   Rotate_Translate)
from ..augmentation.splines import (Spline_Curve_Simplification,
                                   Spline_X_Simplification,
                                   Spline_Y_Perturbations,
                                   Spline_X_Perturbations,
                                   Spline_Smoothing)
from ..augmentation.abc_augmenter import Augmenter, IdentityAugmenter

from sklearn.preprocessing import FunctionTransformer as IdentityTransformer
from sklearn.preprocessing import RobustScaler as RobustNormalVariate
from sklearn.preprocessing import StandardScaler as StandardNormalVariate

from .nirs import (Haar, MultiplicativeScatterCorrection, SavitzkyGolay, Wavelet, msc, savgol, wavelet_transform, LogTransform, FirstDerivative, SecondDerivative, log_transform, first_derivative, second_derivative)
from .scalers import (Derivate, Normalize, SimpleScale, derivate, norml, spl_norml)
from .signal import Baseline, Detrend, Gaussian, baseline, detrend, gaussian
from .features import CropTransformer, ResampleTransformer


__all__ = [
    # Data augmentation
    "Spline_Smoothing",
    "Spline_X_Perturbations",
    "Spline_Y_Perturbations",
    "Spline_X_Simplification",
    "Spline_Curve_Simplification",
    "Rotate_Translate",
    "Random_X_Operation",
    "Augmenter",
    "IdentityAugmenter",

    # Sklearn aliases
    "IdentityTransformer",  # sklearn.preprocessing.FunctionTransformer alias
    "StandardNormalVariate",  # sklearn.preprocessing.StandardScaler alias
    "RobustNormalVariate",  # sklearn.preprocessing.RobustScaler alias

    # NIRS transformations
    "SavitzkyGolay",
    "Haar",
    "MultiplicativeScatterCorrection",
    "Wavelet",
    "savgol",
    "msc",
    "wavelet_transform",
    "LogTransform",
    "FirstDerivative",
    "SecondDerivative",
    "log_transform",
    "first_derivative",
    "second_derivative",

    # Scalers
    "Normalize",
    "Derivate",
    "SimpleScale",
    "norml",
    "derivate",
    "spl_norml",

    # Signal processing
    "Baseline",
    "Detrend",
    "Gaussian",
    "baseline",
    "detrend",
    "gaussian",

    # Features
    "CropTransformer",
    "ResampleTransformer"
]



# """
# Input/Output functionality for SpectroDataset persistence.

# This module provides save/load functionality for complete datasets,
# storing features as numpy files and metadata as parquet files.
# """

# import os
# import json
# from typing import TYPE_CHECKING

# import numpy as np
# import polars as pl

# if TYPE_CHECKING:
#     from .dataset import SpectroDataset


# def save(ds: "SpectroDataset", path: str) -> None:
#     """
#     Save a SpectroDataset to disk.

#     Creates a folder with:
#     - features_src#.npy files (one per source)
#     - targets.parquet
#     - metadata.parquet
#     - predictions.parquet
#     - index.parquet
#     - folds.json

#     Args:
#         ds: SpectroDataset instance to save
#         path: Directory path where to save the dataset
#     """
#     # Create directory
#     os.makedirs(path, exist_ok=True)

#     # Save features as numpy files
#     if ds.features.sources:
#         features_base = os.path.join(path, "features")
#         ds.features.dump_numpy(features_base)

#     # Save index if available
#     if ds.features.index_df is not None:
#         index_path = os.path.join(path, "index.parquet")
#         ds.features.index_df.write_parquet(index_path)    # Save targets if available
#     if ds.targets.sources:
#         # For the new TargetBlock, we need to serialize the sources
#         # For now, let's save the raw target data - this is a simplified implementation
#         # A more complete implementation would serialize all the source metadata
#         import polars as pl
#         target_rows = []
#         for key, source in ds.targets.sources.items():
#             # Extract name and processing from key
#             parts = key.rsplit('_', 1)
#             name = parts[0] if len(parts) > 1 else key
#             processing = parts[1] if len(parts) > 1 else "raw"

#             # Convert source data to table format
#             data = source.get_raw_data()
#             if data.ndim == 1:
#                 data = data.reshape(-1, 1)

#             for i, sample_id in enumerate(source.samples):
#                 target_rows.append({
#                     'sample': int(sample_id),
#                     'targets': data[i].tolist(),
#                     'processing': processing,
#                     'target_type': source.target_type.value,
#                     'name': name
#                 })

#         if target_rows:
#             targets_df = pl.DataFrame(target_rows)
#             targets_path = os.path.join(path, "targets.parquet")
#             targets_df.write_parquet(targets_path)

#     # Save metadata if available
#     if ds.metadata.table is not None:
#         metadata_path = os.path.join(path, "metadata.parquet")
#         ds.metadata.table.write_parquet(metadata_path)

#     # Save predictions if available
#     if ds.predictions.table is not None:
#         predictions_path = os.path.join(path, "predictions.parquet")
#         ds.predictions.table.write_parquet(predictions_path)

#     # Save folds if available
#     if ds.folds.folds:
#         folds_path = os.path.join(path, "folds.json")
#         with open(folds_path, 'w', encoding='utf-8') as f:
#             json.dump(ds.folds.folds, f, indent=2)


# def load(path: str) -> "SpectroDataset":
#     """
#     Load a SpectroDataset from disk using zero-copy memory mapping.

#     Args:
#         path: Directory path containing the saved dataset

#     Returns:
#         Reconstructed SpectroDataset instance

#     Raises:
#         FileNotFoundError: If the directory does not exist
#     """
#     from .dataset import SpectroDataset

#     if not os.path.exists(path):
#         raise FileNotFoundError(f"Dataset directory not found: {path}")

#     if not os.path.isdir(path):
#         raise NotADirectoryError(f"Path is not a directory: {path}")

#     ds = SpectroDataset()

#     # Load features with memory mapping
#     features_base = os.path.join(path, "features")
#     if os.path.exists(f"{features_base}_src0.npy"):
#         ds.features.load_numpy(features_base, mmap_mode="r")    # Load index if available
#     index_path = os.path.join(path, "index.parquet")
#     if os.path.exists(index_path):
#         ds.features.index_df = pl.read_parquet(index_path)

#     # Load targets if available
#     targets_path = os.path.join(path, "targets.parquet")
#     if os.path.exists(targets_path):
#         targets_df = pl.read_parquet(targets_path)

#         # Group by name and processing to reconstruct sources
#         if 'name' in targets_df.columns and 'target_type' in targets_df.columns:
#             # New format with explicit target metadata
#             for (name, processing, target_type), group in targets_df.group_by(['name', 'processing', 'target_type']):
#                 samples = group['sample'].to_numpy()
#                 targets_list = group['targets'].to_list()
#                 target_data = np.array(targets_list, dtype=np.float32)

#                 # Convert to strings for type safety
#                 name_str = str(name)
#                 processing_str = str(processing)
#                 target_type_str = str(target_type)

#                 if target_type_str == 'regression':
#                     ds.targets.add_regression_targets(name_str, target_data, samples, processing_str)
#                 elif target_type_str == 'classification':
#                     ds.targets.add_classification_targets(name_str, target_data.flatten().astype(int), samples, processing_str)
#                 elif target_type_str == 'multilabel':
#                     ds.targets.add_multilabel_targets(name_str, target_data, samples, processing_str)
#         else:
#             # Legacy format - assume regression
#             samples = targets_df['sample'].to_numpy()
#             targets_list = targets_df['targets'].to_list()
#             target_data = np.array(targets_list, dtype=np.float32)
#             processing = targets_df['processing'].to_list()[0] if 'processing' in targets_df.columns else "raw"
#             ds.targets.add_regression_targets("target", target_data, samples, processing)

#     # Load metadata if available
#     metadata_path = os.path.join(path, "metadata.parquet")
#     if os.path.exists(metadata_path):
#         ds.metadata.table = pl.read_parquet(metadata_path)

#     # Load predictions if available
#     predictions_path = os.path.join(path, "predictions.parquet")
#     if os.path.exists(predictions_path):
#         ds.predictions.table = pl.read_parquet(predictions_path)

#     # Load folds if available
#     folds_path = os.path.join(path, "folds.json")
#     if os.path.exists(folds_path):
#         with open(folds_path, 'r', encoding='utf-8') as f:
#             ds.folds.folds = json.load(f)

#     return ds



from typing import Any, Dict, List, Tuple, Optional, TYPE_CHECKING

from sklearn.base import TransformerMixin

from nirs4all.controllers.controller import OperatorController
from nirs4all.controllers.registry import register_controller

if TYPE_CHECKING:
    from nirs4all.pipeline.runner import PipelineRunner
    from nirs4all.spectra.spectra_dataset import SpectroDataset
import copy


@register_controller
class FeatureAugmentationController(OperatorController):
    priority = 10

    @classmethod
    def matches(cls, step: Any, operator: Any, keyword: str) -> bool:
        return keyword == "feature_augmentation"

    @classmethod
    def use_multi_source(cls) -> bool:
        """Check if the operator supports multi-source datasets."""
        return True

    @classmethod
    def supports_prediction_mode(cls) -> bool:
        """Feature augmentation should NOT execute during prediction mode - transformations are already applied and saved."""
        return True

    def execute(  # TODO reup parralelization
        self,
        step: Any,
        operator: Any,
        dataset: 'SpectroDataset',
        context: Dict[str, Any],
        runner: 'PipelineRunner',
        source: int = -1,
        mode: str = "train",
        loaded_binaries: Optional[List[Tuple[str, Any]]] = None,
        prediction_store: Optional[Any] = None
    ) -> Tuple[Dict[str, Any], List[Tuple[str, bytes]]]:
        # print(f"Executing feature augmentation for step: {step}, keyword: {context.get('keyword', '')}, source: {source}, mode: {mode}")

        try:
            initial_context = copy.deepcopy(context)
            # Faire une deepcopy à chaque utilisation pour éviter les modifications
            original_source_processings = copy.deepcopy(initial_context["processing"])

            for i, operation in enumerate(step["feature_augmentation"]):
                # Recréer source_processings à chaque itération pour éviter les mutations
                source_processings = copy.deepcopy(original_source_processings)
                local_context = copy.deepcopy(initial_context)
                # print(f"Applying feature augmentation operation {i + 1}/{len(step['feature_augmentation'])}: {operation}")
                # if i == 0 and operation is None:
                #     print("Skipping no-op feature augmentation")
                #     continue
                # if i > 0:
                local_context["add_feature"] = True

                # Assigner une nouvelle copie à chaque fois
                local_context["processing"] = copy.deepcopy(source_processings)
                runner.run_step(operation, dataset, local_context, prediction_store, is_substep=True, propagated_binaries=loaded_binaries)

            context["processing"] = []
            for sdx in range(dataset.n_sources):
                processing_ids = dataset.features_processings(sdx)
                context["processing"].append(processing_ids)
            return context, []

        except Exception as e:
            print(f"❌ Error applying feature augmentation: {e}")
            raise


"""Inception 1D_CNN Models in Tensorflow-Keras.
References -
Inception_v1 (GoogLeNet): https://arxiv.org/abs/1409.4842 [Going Deeper with Convolutions]
Inception_v2: http://arxiv.org/abs/1512.00567 [Rethinking the Inception Architecture for Computer Vision]
Inception_v3: http://arxiv.org/abs/1512.00567 [Rethinking the Inception Architecture for Computer Vision]
Inception_v4: https://arxiv.org/abs/1602.07261 [Inception-v4, Inception-ResNet and the Impact of Residual Connections on Learning]
"""


import tensorflow as tf


def Conv_1D_Block(x, model_width, kernel, strides=1, padding="same"):
    # 1D Convolutional Block with BatchNormalization
    x = tf.keras.layers.Conv1D(model_width, kernel, strides=strides, padding=padding, kernel_initializer="he_normal")(x)
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.layers.Activation('relu')(x)

    return x


def classifier(inputs, class_number):
    # Construct the Classifier Group
    # inputs       : input vector
    # class_number : number of output classes
    out = tf.keras.layers.Dense(class_number, activation='softmax')(inputs)
    return out


def regressor(inputs, feature_number):
    # Construct the Regressor Group
    # inputs         : input vector
    # feature_number : number of output features
    out = tf.keras.layers.Dense(feature_number, activation='linear')(inputs)
    return out


def Inceptionv1_Module(inputs, filterB1_1, filterB2_1, filterB2_2, filterB3_1, filterB3_2, filterB4_1, i):
    # Inception Block i
    branch1x1 = Conv_1D_Block(inputs, filterB1_1, 1, padding='valid')

    branch3x3 = Conv_1D_Block(inputs, filterB2_1, 1, padding='valid')
    branch3x3 = Conv_1D_Block(branch3x3, filterB2_2, 3)

    branch5x5 = Conv_1D_Block(inputs, filterB3_1, 1, padding='valid')
    branch5x5 = Conv_1D_Block(branch5x5, filterB3_2, 5)

    branch_pool = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inputs)
    branch_pool = Conv_1D_Block(branch_pool, filterB4_1, 1)
    out = tf.keras.layers.concatenate([branch1x1, branch3x3, branch5x5, branch_pool], axis=-1, name='Inception_Block_'+str(i))

    return out


def Inceptionv2_Module(inputs, filterB1_1, filterB2_1, filterB2_2, filterB3_1, filterB3_2, filterB3_3, filterB4_1, i):
    # Inception Block i
    branch1x1 = Conv_1D_Block(inputs, filterB1_1, 1)

    branch3x3 = Conv_1D_Block(inputs, filterB2_1, 1)
    branch3x3 = Conv_1D_Block(branch3x3, filterB2_2, 3)

    branch3x3dbl = Conv_1D_Block(inputs, filterB3_1, 1)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_2, 3)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_3, 3)

    branch_pool = tf.keras.layers.AveragePooling1D(pool_size=3, strides=1, padding='same')(inputs)
    branch_pool = Conv_1D_Block(branch_pool, filterB4_1, 1)

    out = tf.keras.layers.concatenate([branch1x1, branch3x3, branch3x3dbl, branch_pool], axis=-1, name='Inception_Block_'+str(i))

    return out


def Inception_Module_A(inputs, filterB1_1, filterB2_1, filterB2_2, filterB3_1, filterB3_2, filterB3_3, filterB4_1, i):
    # Inception Block i
    branch1x1 = Conv_1D_Block(inputs, filterB1_1, 1)

    branch3x3 = Conv_1D_Block(inputs, filterB2_1, 1)
    branch3x3 = Conv_1D_Block(branch3x3, filterB2_2, 5)

    branch3x3dbl = Conv_1D_Block(inputs, filterB3_1, 1)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_2, 3)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_3, 3)

    branch_pool = tf.keras.layers.AveragePooling1D(pool_size=3, strides=1, padding='same')(inputs)
    branch_pool = Conv_1D_Block(branch_pool, filterB4_1, 1)

    out = tf.keras.layers.concatenate([branch1x1, branch3x3, branch3x3dbl, branch_pool], axis=-1, name='Inception_Block_A'+str(i))

    return out


def Inception_Module_B(inputs, filterB1_1, filterB2_1, filterB2_2, filterB3_1, filterB3_2, filterB3_3, filterB4_1, i):
    # Inception Block i
    branch1x1 = Conv_1D_Block(inputs, filterB1_1, 1)

    branch3x3 = Conv_1D_Block(inputs, filterB2_1, 1)
    branch3x3 = Conv_1D_Block(branch3x3, filterB2_2, 7)

    branch3x3dbl = Conv_1D_Block(inputs, filterB3_1, 1)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_2, 7)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_3, 7)

    branch_pool = tf.keras.layers.AveragePooling1D(pool_size=3, strides=1, padding='same')(inputs)
    branch_pool = Conv_1D_Block(branch_pool, filterB4_1, 1)

    out = tf.keras.layers.concatenate([branch1x1, branch3x3, branch3x3dbl, branch_pool], axis=-1, name='Inception_Block_B'+str(i))

    return out


def Inception_Module_C(inputs, filterB1_1, filterB2_1, filterB2_2, filterB3_1, filterB3_2, filterB3_3, filterB4_1, i):
    # Inception Block i
    branch1x1 = Conv_1D_Block(inputs, filterB1_1, 1)

    branch3x3 = Conv_1D_Block(inputs, filterB2_1, 1)
    branch3x3 = Conv_1D_Block(branch3x3, filterB2_2, 3)

    branch3x3dbl = Conv_1D_Block(inputs, filterB3_1, 1)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_2, 3)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB3_3, 3)

    branch_pool = tf.keras.layers.AveragePooling1D(pool_size=3, strides=1, padding='same')(inputs)
    branch_pool = Conv_1D_Block(branch_pool, filterB4_1, 1)

    out = tf.keras.layers.concatenate([branch1x1, branch3x3, branch3x3dbl, branch_pool], axis=-1, name='Inception_Block_C'+str(i))

    return out


def Reduction_Block_A(inputs, filterB1_1, filterB1_2, filterB2_1, filterB2_2, filterB2_3, i):
    # Reduction Block A (i)
    branch3x3 = Conv_1D_Block(inputs, filterB1_1, 1)
    branch3x3 = Conv_1D_Block(branch3x3, filterB1_2, 3, strides=2)

    branch3x3dbl = Conv_1D_Block(inputs, filterB2_1, 1)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB2_2, 3)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB2_3, 3, strides=2)

    branch_pool = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inputs)
    out = tf.keras.layers.concatenate([branch3x3, branch3x3dbl, branch_pool], axis=-1, name='Reduction_Block_'+str(i))

    return out


def Reduction_Block_B(inputs, filterB1_1, filterB1_2, filterB2_1, filterB2_2, filterB2_3, i):
    # Reduction Block B (i)
    branch3x3 = Conv_1D_Block(inputs, filterB1_1, 1)
    branch3x3 = Conv_1D_Block(branch3x3, filterB1_2, 3, strides=2)

    branch3x3dbl = Conv_1D_Block(inputs, filterB2_1, 1)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB2_2, 7)
    branch3x3dbl = Conv_1D_Block(branch3x3dbl, filterB2_3, 3, strides=2)

    branch_pool = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inputs)
    out = tf.keras.layers.concatenate([branch3x3, branch3x3dbl, branch_pool], axis=-1, name='Reduction_Block_'+str(i))

    return out


class Inception:
    def __init__(self, length, num_channel, num_filters, problem_type='Regression',
                 output_nums=1, pooling='avg', dropout_rate=False, auxilliary_outputs=False):
        # length: Input Signal Length
        # model_depth: Depth of the Model
        # model_width: Width of the Model
        # kernel_size: Kernel or Filter Size of the Input Convolutional Layer
        # num_channel: Number of Channels of the Input Predictor Signals
        # problem_type: Regression or Classification
        # output_nums: Number of Output Classes in Classification mode and output features in Regression mode
        # pooling: Choose either 'max' for MaxPooling or 'avg' for Averagepooling
        # dropout_rate: If turned on, some layers will be dropped out randomly based on the selected proportion
        # auxilliary_outputs: Two extra Auxullary outputs for the Inception models, acting like Deep Supervision
        self.length = length
        self.num_channel = num_channel
        self.num_filters = num_filters
        self.problem_type = problem_type
        self.output_nums = output_nums
        self.pooling = pooling
        self.dropout_rate = dropout_rate
        self.auxilliary_outputs = auxilliary_outputs

    def MLP(self, x):
        if self.pooling == 'avg':
            x = tf.keras.layers.GlobalAveragePooling1D()(x)
        elif self.pooling == 'max':
            x = tf.keras.layers.GlobalMaxPooling1D()(x)
        # Final Dense Outputting Layer for the outputs
        x = tf.keras.layers.Flatten()(x)
        if self.dropout_rate:
            x = tf.keras.layers.Dropout(self.dropout_rate)(x)
        outputs = tf.keras.layers.Dense(self.output_nums, activation='linear')(x)
        if self.problem_type == 'Classification':
            outputs = tf.keras.layers.Dense(self.output_nums, activation='softmax')(x)

        return outputs

    def Inception_v1(self):
        inputs = tf.keras.Input((self.length, self.num_channel))  # The input tensor
        # Stem
        x = Conv_1D_Block(inputs, self.num_filters, 7, strides=2)
        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)
        x = Conv_1D_Block(x, self.num_filters, 1, padding='valid')
        x = Conv_1D_Block(x, self.num_filters * 3, 3)
        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)

        x = Inceptionv1_Module(x, 64, 96, 128, 16, 32, 32, 1)  # Inception Block 1
        x = Inceptionv1_Module(x, 128, 128, 192, 32, 96, 64, 2)  # Inception Block 2

        aux_output_0 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 0
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 64, 1)
            aux_output_0 = self.MLP(aux_conv)

        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)
        x = Inceptionv1_Module(x, 192, 96, 208, 16, 48, 64, 3)  # Inception Block 3
        x = Inceptionv1_Module(x, 160, 112, 224, 24, 64, 64, 4)  # Inception Block 4
        x = Inceptionv1_Module(x, 128, 128, 256, 24, 64, 64, 5)  # Inception Block 5
        x = Inceptionv1_Module(x, 112, 144, 288, 32, 64, 64, 6)  # Inception Block 6
        x = Inceptionv1_Module(x, 256, 160, 320, 32, 128, 128, 7)  # Inception Block 7

        aux_output_1 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 1
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 64, 1)
            aux_output_1 = self.MLP(aux_conv)

        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)
        x = Inceptionv1_Module(x, 256, 160, 320, 32, 128, 128, 8)  # Inception Block 8
        x = Inceptionv1_Module(x, 384, 192, 384, 48, 128, 128, 9)  # Inception Block 9

        # Final Dense MLP Layer for the outputs
        final_output = self.MLP(x)
        # Create model.
        model = tf.keras.Model(inputs, final_output, name='Inception_v3')
        if self.auxilliary_outputs:
            model = tf.keras.Model(inputs, outputs=[final_output, aux_output_0, aux_output_1], name='Inception_v1')

        return model

    def Inception_v2(self):
        inputs = tf.keras.Input((self.length, self.num_channel))  # The input tensor
        # Stem: 56 x 64
        x = tf.keras.layers.SeparableConv1D(self.num_filters, kernel_size=7, strides=2, depth_multiplier=1, padding='same')(inputs)
        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)
        x = Conv_1D_Block(x, self.num_filters * 2, 1, padding='valid')
        x = Conv_1D_Block(x, self.num_filters * 6, 3, padding='valid')
        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)

        x = Inceptionv2_Module(x, 64, 64, 64, 64, 96, 96, 32, 1)  # Inception Block 1: 28 x 192
        x = Inceptionv2_Module(x, 64, 64, 96, 64, 96, 96, 64, 2)  # Inception Block 2: 28 x 256

        aux_output_0 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 0
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 64, 1)
            aux_output_0 = self.MLP(aux_conv)

        x = Reduction_Block_A(x, 128, 160, 64, 96, 96, 1)  # Reduction Block 1: 28 x 320
        x = Inceptionv2_Module(x, 224, 64, 96, 96, 128, 128, 128, 3)  # Inception Block 3: 14 x 576
        x = Inceptionv2_Module(x, 192, 96, 128, 96, 128, 128, 128, 4)  # Inception Block 4: 14 x 576
        x = Inceptionv2_Module(x, 160, 128, 160, 128, 160, 160, 96, 5)  # Inception Block 5: 14 x 576
        x = Inceptionv2_Module(x, 96, 128, 192, 160, 192, 192, 96, 6)  # Inception Block 6: 14 x 576

        aux_output_1 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 1
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 192, 1)
            aux_output_1 = self.MLP(aux_conv)

        x = Reduction_Block_A(x, 128, 192, 192, 256, 256, 2)  # Reduction Block 2: 14 x 576
        x = Inceptionv2_Module(x, 352, 192, 320, 160, 224, 224, 128, 7)  # Inception Block 7: 7 x 1024
        x = Inceptionv2_Module(x, 352, 192, 320, 192, 224, 224, 128, 8)  # Inception Block 8: 7 x 1024

        # Final Dense MLP Layer for the outputs
        final_output = self.MLP(x)
        # Create model.
        model = tf.keras.Model(inputs, final_output, name='Inception_v3')
        if self.auxilliary_outputs:
            model = tf.keras.Model(inputs, outputs=[final_output, aux_output_0, aux_output_1], name='Inception_v2')

        return model

    def Inception_v3(self):
        inputs = tf.keras.Input((self.length, self.num_channel))  # The input tensor
        # Stem
        x = Conv_1D_Block(inputs, self.num_filters, 3, strides=2, padding='valid')
        x = Conv_1D_Block(x, self.num_filters, 3, padding='valid')
        x = Conv_1D_Block(x, self.num_filters * 2, 3)
        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)

        x = Conv_1D_Block(x, self.num_filters * 2.5, 1, padding='valid')
        x = Conv_1D_Block(x, self.num_filters * 6, 3, padding='valid')
        x = tf.keras.layers.MaxPooling1D(3, strides=2)(x)

        # 3x Inception-A Blocks
        x = Inception_Module_A(x, 64, 48, 64, 64, 96, 96, 32, 1)  # Inception-A Block 1: 35 x 256
        x = Inception_Module_A(x, 64, 48, 64, 64, 96, 96, 64, 2)  # Inception-A Block 2: 35 x 256
        x = Inception_Module_A(x, 64, 48, 64, 64, 96, 96, 64, 3)  # Inception-A Block 3: 35 x 256

        aux_output_0 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 0
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 64, 1)
            aux_output_0 = self.MLP(aux_conv)

        x = Reduction_Block_A(x, 64, 384, 64, 96, 96, 1)  # Reduction Block 1: 17 x 768

        # 4x Inception-B Blocks
        x = Inception_Module_B(x, 192, 128, 192, 128, 128, 192, 192, 1)  # Inception-B Block 1: 17 x 768
        x = Inception_Module_B(x, 192, 160, 192, 160, 160, 192, 192, 2)  # Inception-B Block 2: 17 x 768
        x = Inception_Module_B(x, 192, 160, 192, 160, 160, 192, 192, 3)  # Inception-B Block 3: 17 x 768
        x = Inception_Module_B(x, 192, 192, 192, 192, 192, 192, 192, 4)  # Inception-B Block 4: 17 x 768

        aux_output_1 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 1
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 192, 1)
            aux_output_1 = self.MLP(aux_conv)

        x = Reduction_Block_B(x, 192, 320, 192, 192, 192, 2)  # Reduction Block 2: 8 x 1280

        # 2x Inception-C Blocks: 8 x 2048
        x = Inception_Module_C(x, 320, 384, 384, 448, 384, 384, 192, 1)  # Inception-C Block 1: 8 x 2048
        x = Inception_Module_C(x, 320, 384, 384, 448, 384, 384, 192, 2)  # Inception-C Block 2: 8 x 2048

        # Final Dense MLP Layer for the outputs
        final_output = self.MLP(x)
        # Create model.
        model = tf.keras.Model(inputs, final_output, name='Inception_v3')
        if self.auxilliary_outputs:
            model = tf.keras.Model(inputs, outputs=[final_output, aux_output_0, aux_output_1], name='Inception_v3')

        return model

    def Inception_v4(self):
        inputs = tf.keras.Input((self.length, self.num_channel))  # The input tensor
        # Stem
        x = Conv_1D_Block(inputs, 32, 3, strides=2, padding='valid')
        x = Conv_1D_Block(x, 32, 3, padding='valid')
        x = Conv_1D_Block(x, 64, 3)

        branch1 = Conv_1D_Block(x, 96, 3, strides=2, padding='valid')
        branch2 = tf.keras.layers.MaxPooling1D(3, strides=2)(x)
        x = tf.keras.layers.concatenate([branch1, branch2], axis=-1)

        branch1 = Conv_1D_Block(x, 64, 1)
        branch1 = Conv_1D_Block(branch1, 96, 3, padding='valid')
        branch2 = Conv_1D_Block(x, 64, 1)
        branch2 = Conv_1D_Block(branch2, 64, 7)
        branch2 = Conv_1D_Block(branch2, 96, 3, padding='valid')
        x = tf.keras.layers.concatenate([branch1, branch2], axis=-1)

        branch1 = Conv_1D_Block(x, 192, 3, padding='valid')
        branch2 = tf.keras.layers.MaxPooling1D(3, strides=2)(x)
        x = tf.keras.layers.concatenate([branch1, branch2], axis=1)

        # 4x Inception-A Blocks - 35 x 256
        for i in range(4):
            x = Inception_Module_A(x, 96, 64, 96, 64, 96, 96, 96, i)

        aux_output_0 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 0
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 96, 1)
            aux_output_0 = self.MLP(aux_conv)

        x = Reduction_Block_A(x, 64, 384, 192, 224, 256, 1)  # Reduction Block 1: 17 x 768

        # 7x Inception-B Blocks - 17 x 768
        for i in range(7):
            x = Inception_Module_B(x, 384, 192, 256, 192, 224, 256, 128, i)

        aux_output_1 = []
        if self.auxilliary_outputs:
            # Auxilliary Output 1
            aux_pool = tf.keras.layers.AveragePooling1D(pool_size=5, strides=3, padding='valid')(x)
            aux_conv = Conv_1D_Block(aux_pool, 128, 1)
            aux_output_1 = self.MLP(aux_conv)

        x = Reduction_Block_B(x, 192, 192, 256, 320, 320, 2)  # Reduction Block 2: 8 x 1280

        # 3x Inception-C Blocks: 8 x 2048
        for i in range(3):
            x = Inception_Module_C(x, 256, 384, 512, 384, 512, 512, 256, i)

        # Final Dense MLP Layer for the outputs
        final_output = self.MLP(x)
        # Create model.
        model = tf.keras.Model(inputs, final_output, name='Inception_v4')
        if self.auxilliary_outputs:
            model = tf.keras.layers.Model(inputs, outputs=[final_output, aux_output_0, aux_output_1], name='Inception_v4')

        return model



import numpy as np
# import random
import operator
# from ast import operator

from .abc_augmenter import Augmenter


def angle_p(x, xI, yI, p1, p2):
    """
    Helper function to calculate the angle for rotation and translation.

    Parameters
    ----------
    x : float
        Input value.
    xI : float
        Reference point.
    yI : float
        Initial value.
    p1 : float
        Slope 1.
    p2 : float
        Slope 2.

    Returns
    -------
    float
        Calculated angle.
    """
    mask = x <= xI
    return np.where(mask, p1 * (x - xI) + yI, p2 * (x - xI) + yI)


v_angle_p = np.vectorize(angle_p)


class Rotate_Translate(Augmenter):
    """
    Class for rotating and translating data augmentation.

    Parameters
    ----------
    apply_on : str, optional
        Apply augmentation on "samples" or "global" data. Default is "samples".
    random_state : int or None, optional
        Random seed for reproducibility. Default is None.
    copy : bool, optional
        If True, creates a copy of the input data. Default is True.
    p_range : int, optional
        Range for generating random slope values. Default is 2.
    y_factor : int, optional
        Scaling factor for the initial value. Default is 3.
    """

    def __init__(self, apply_on="samples", random_state=None, *, copy=True, p_range=2, y_factor=3):
        self.p_range = p_range
        self.y_factor = y_factor
        super().__init__(apply_on, random_state, copy=copy)

    def augment(self, X, apply_on="samples"):
        """
        Augment the data by rotating and translating the signal.

        Parameters
        ----------
        X : ndarray
            Input data to be augmented.
        apply_on : str, optional
            Apply augmentation on "samples" or "global" data. Default is "samples".

        Returns
        -------
        ndarray
            Augmented data.
        """
        # def deformation(x):
        #     x_range = np.linspace(0, 1, x.shape[-1])
        #     p2 = self.random_gen.uniform(-self.p_range/5, self.p_range/5)
        #     p1 = self.random_gen.uniform(-self.p_range/5, self.p_range/5)
        #     xI = self.random_gen.uniform(0, 1)
        #     yI = self.random_gen.uniform(0, np.max(x) / self.y_factor)
        #     distor = v_angle_p(x_range, xI, yI, p1, p2)
        #     return distor

        def deformation(x):
            x_range = np.linspace(0, 1, x.shape[-1])
            p2 = self.random_gen.uniform(-self.p_range/5, self.p_range/5)
            p1 = self.random_gen.uniform(-self.p_range/5, self.p_range/5)
            xI = self.random_gen.uniform(0, 1)
            yI = self.random_gen.uniform(0, max(0, np.max(x) / self.y_factor))  # Ensure yI range is valid
            distor = v_angle_p(x_range, xI, yI, p1, p2)
            return distor


        if apply_on == "global":
            increment = deformation(X) * np.std(X)
        else:
            increment = np.array([deformation(x) * np.std(x) for x in X])

        new_X = X + increment

        return new_X


class Random_X_Operation(Augmenter):
    """
    Class for applying random operation on data augmentation.

    Parameters
    ----------
    apply_on : str, optional
        Apply augmentation on "features" or "samples" data. Default is "features".
    random_state : int or None, optional
        Random seed for reproducibility. Default is None.
    copy : bool, optional
        If True, creates a copy of the input data. Default is True.
    operator_func : function, optional
        Operator function to be applied. Default is operator.mul.
    operator_range : tuple, optional
        Range for generating random values for the operator. Default is (0.97, 1.03).
    """

    def __init__(self, apply_on="global", random_state=None, *, copy=True, operator_func=operator.mul, operator_range=(0.97, 1.03)):
        self.operator_func = operator_func
        self.operator_range = operator_range
        super().__init__(apply_on, random_state, copy=copy)

    def augment(self, X, apply_on="global"):
        """
        Augment the data by applying random operation.

        Parameters
        ----------
        X : ndarray
            Input data to be augmented.
        apply_on : str, optional
            Apply augmentation on "features" or "samples" data. Default is "features".

        Returns
        -------
        ndarray
            Augmented data.
        """
        min_val = self.operator_range[0]
        interval = self.operator_range[1] - self.operator_range[0]

        if apply_on == "global":
            increment = self.random_gen.random(X.shape[-1]) * interval + min_val
        else:
            increment = self.random_gen.random(X.shape) * interval + min_val


        new_X = self.operator_func(X, increment)
        # Clip the augmented data within the float32 range
        new_X = np.clip(new_X, -np.finfo(np.float32).max, np.finfo(np.float32).max)

        # Log the min and max values to help debug any potential overflow
        # print("Augmented Data Range:", new_X.min(), new_X.max())

        return new_X



# Fix: Add list handling to expand_spec function
from itertools import product, combinations, permutations
from collections.abc import Mapping
from math import comb, factorial
import random

def expand_spec(node):
    # NEW: Handle lists by expanding each element and taking the product
    if isinstance(node, list):
        if not node:
            return [[]]  # Empty list -> single empty result

        # Special case: if there's only one element in the list, check if we should unwrap
        if len(node) == 1:
            element_result = expand_spec(node[0])

            # If the result contains lists (combinations), return directly to prevent extra wrapping
            # If the result contains scalars, we still need to wrap them properly
            if element_result and isinstance(element_result[0], list):
                return element_result
            # Otherwise, fall through to normal processing

        # Expand each element in the list
        expanded_elements = [expand_spec(element) for element in node]

        # Take Cartesian product of all expansions
        results = []
        for combo in product(*expanded_elements):
            results.append(list(combo))  # Convert tuple to list
        return results

    # Rest of the logic remains the same for dictionaries
    if not isinstance(node, Mapping):
        return [node]

    # NEW: Handle _range_ node for numeric sequences
    if set(node.keys()).issubset({"_range_", "count"}):
        range_spec = node["_range_"]
        count = node.get("count", None)

        # Generate numeric range based on specification
        range_values = _generate_range(range_spec)

        # Apply count limit if specified
        if count is not None and len(range_values) > count:
            range_values = random.sample(range_values, count)

        return range_values

    # Case 1: pure OR node (with optional size, count)
    if set(node.keys()).issubset({"_or_", "size", "count"}):
        choices = node["_or_"]
        size = node.get("size", None)
        count = node.get("count", None)

        if size is not None:
            # NEW: Check for second-order array notation [outer, inner]
            if isinstance(size, list) and len(size) == 2:
                results = _handle_nested_combinations(choices, size)
                # Apply count limit if specified
                if count is not None and len(results) > count:
                    results = random.sample(results, count)
                return results

            # Apply size constraints first, then expand
            out = []

            # Handle tuple size (from, to) or single size
            if isinstance(size, tuple) and len(size) == 2:
                from_size, to_size = size
                # Generate combinations for all sizes from from_size to to_size (inclusive)
                for s in range(from_size, to_size + 1):
                    if s > len(choices):
                        continue
                    for combo in combinations(choices, s):
                        # Expand this specific combination and return as individual elements
                        combo_results = _expand_combination(combo)
                        out.extend(combo_results)
            else:
                # Single size value
                if size <= len(choices):
                    for combo in combinations(choices, size):
                        # Expand this specific combination and return as individual elements
                        combo_results = _expand_combination(combo)
                        out.extend(combo_results)

            # Apply count limit if specified
            if count is not None and len(out) > count:
                out = random.sample(out, count)

            return out
        else:
            # Original behavior: expand all choices
            out = []
            for choice in choices:
                out.extend(expand_spec(choice))

            # Apply count limit if specified
            if count is not None and len(out) > count:
                out = random.sample(out, count)

            return out

    # Case 2: dict that also contains "_or_" -> branch and merge
    if "_or_" in node:
        # Extract size and count if present
        size = node.get("size", None)
        count = node.get("count", None)
        base = {k: v for k, v in node.items() if k not in ["_or_", "size", "count"]}
        base_expanded = expand_spec(base)            # list[dict]

        # Create a temporary or node with size and count
        or_node = {"_or_": node["_or_"]}
        if size is not None:
            or_node["size"] = size
        if count is not None:
            or_node["count"] = count

        choice_expanded = expand_spec(or_node)  # list[dict or scalar]
        results = []
        for b in base_expanded:
            for c in choice_expanded:
                if isinstance(c, Mapping):
                    merged = {**b, **c}
                    results.append(merged)
                else:
                    # Scalar choices only make sense as values under a key, not top-level merges
                    raise ValueError("Top-level 'or' choices must be dicts.")
        return results

    # Case 3: normal dict -> product over keys
    keys, options = zip(*[(k, _expand_value(v)) for k, v in node.items()]) if node else ([], [])
    if not keys:
        return [{}]
    out = []
    for combo in product(*options):
        d = {}
        for k, v in zip(keys, combo):
            d[k] = v
        out.append(d)
    return out

def count_combinations(node):
    """Calculate total number of combinations without generating them.
    Returns the count of results that expand_spec would produce.
    """
    # Handle lists by taking product of counts
    if isinstance(node, list):
        if not node:
            return 1  # Empty list -> single empty result

        total_count = 1
        for element in node:
            total_count *= count_combinations(element)
        return total_count

    # Scalars return 1
    if not isinstance(node, Mapping):
        return 1

    # NEW: Handle _range_ node for counting numeric sequences
    if set(node.keys()).issubset({"_range_", "count"}):
        range_spec = node["_range_"]
        count = node.get("count", None)

        # Calculate range size
        range_size = _count_range(range_spec)

        # Apply count limit if specified
        if count is not None:
            return min(count, range_size)
        return range_size

    # Case 1: pure OR node (with optional size, count)
    if set(node.keys()).issubset({"_or_", "size", "count"}):
        choices = node["_or_"]
        size = node.get("size", None)
        count = node.get("count", None)

        if size is not None:
            # Check for second-order array notation [outer, inner]
            if isinstance(size, list) and len(size) == 2:
                total_count = _count_nested_combinations(choices, size)
                # Apply count limit if specified
                if count is not None:
                    return min(count, total_count)
                return total_count

            # Handle tuple size (from, to) or single size
            total_count = 0
            if isinstance(size, tuple) and len(size) == 2:
                from_size, to_size = size
                for s in range(from_size, to_size + 1):
                    if s <= len(choices):
                        total_count += comb(len(choices), s)
            else:
                # Single size value
                if size <= len(choices):
                    total_count = comb(len(choices), size)

            # Apply count limit if specified
            if count is not None:
                return min(count, total_count)
            return total_count
        else:
            # All choices - count each choice's expansions
            total_count = 0
            for choice in choices:
                total_count += count_combinations(choice)

            # Apply count limit if specified
            if count is not None:
                return min(count, total_count)
            return total_count

    # Case 2: dict that also contains "_or_" -> branch and merge
    if "_or_" in node:
        size = node.get("size", None)
        count = node.get("count", None)
        base = {k: v for k, v in node.items() if k not in ["_or_", "size", "count"]}

        base_count = count_combinations(base)

        # Create temporary or node
        or_node = {"_or_": node["_or_"]}
        if size is not None:
            or_node["size"] = size
        if count is not None:
            or_node["count"] = count

        choice_count = count_combinations(or_node)

        return base_count * choice_count

    # Case 3: normal dict -> product over keys
    if not node:
        return 1

    total_count = 1
    for k, v in node.items():
        total_count *= _count_value(v)

    return total_count

def _handle_nested_combinations(choices, nested_size):
    """Handle second-order combinations with array syntax [outer, inner]
    outer and inner can be int or tuple (from, to)
    For second-order: inner elements use permutations (order matters within sub-arrays)
    For outer selection: uses combinations (order doesn't matter for selecting which sub-arrays)
    """
    outer_size, inner_size = nested_size

    # Handle inner size - can be int or tuple
    if isinstance(inner_size, int):
        inner_from, inner_to = inner_size, inner_size
    else:  # tuple (from, to)
        inner_from, inner_to = inner_size

    # Handle outer size - can be int or tuple
    if isinstance(outer_size, int):
        outer_from, outer_to = outer_size, outer_size
    else:  # tuple (from, to)
        outer_from, outer_to = outer_size

    # Step 1: Generate all possible inner arrangements (permutations - order matters!)
    inner_arrangements = []
    for inner_s in range(inner_from, inner_to + 1):
        if inner_s > len(choices):
            continue
        # Use permutations for inner elements (order matters within sub-arrays)
        for perm in permutations(choices, inner_s):
            # Each inner arrangement is a simple list
            if len(perm) == 1:
                # Single element - don't wrap in extra list
                inner_arrangements.append(perm[0])
            else:
                # Multiple elements - keep as list
                inner_arrangements.append(list(perm))

    # Step 2: Select outer combinations (combinations - order doesn't matter for selection)
    out = []
    for outer_s in range(outer_from, outer_to + 1):
        if outer_s > len(inner_arrangements):
            continue
        # Use combinations for outer selection (order doesn't matter for which sub-arrays to pick)
        for outer_combo in combinations(inner_arrangements, outer_s):
            # Convert tuple to list and add to results
            out.append(list(outer_combo))

    return out

def _count_nested_combinations(choices, nested_size):
    """Count second-order combinations without generating them."""
    outer_size, inner_size = nested_size

    # Handle inner size - can be int or tuple
    if isinstance(inner_size, int):
        inner_from, inner_to = inner_size, inner_size
    else:
        inner_from, inner_to = inner_size

    # Handle outer size - can be int or tuple
    if isinstance(outer_size, int):
        outer_from, outer_to = outer_size, outer_size
    else:
        outer_from, outer_to = outer_size

    # Count inner arrangements (permutations)
    total_inner_arrangements = 0
    n_choices = len(choices)

    for inner_s in range(inner_from, inner_to + 1):
        if inner_s <= n_choices:
            # Permutations: P(n,k) = n!/(n-k)!
            perms = factorial(n_choices) // factorial(n_choices - inner_s) if inner_s <= n_choices else 0
            total_inner_arrangements += perms

    # Count outer combinations from inner arrangements
    total_count = 0
    for outer_s in range(outer_from, outer_to + 1):
        if outer_s <= total_inner_arrangements:
            total_count += comb(total_inner_arrangements, outer_s)

    return total_count

def _expand_combination(combo):
    """Expand a specific combination of choices by taking their cartesian product."""
    expanded_choices = []
    for choice in combo:
        expanded_choices.append(expand_spec(choice))

    # Take cartesian product
    results = []
    for expanded_combo in product(*expanded_choices):
        # Return the combination as a single list (flatten one level)
        results.append(list(expanded_combo))

    return results

def _expand_value(v):
    # Value position returns a list of *values* (scalars or dicts)
    if isinstance(v, Mapping) and ("_or_" in v.keys()):
        # Value-level OR can yield scalars or dicts as values (with optional size, count)
        choices = v["_or_"]
        size = v.get("size", None)
        count = v.get("count", None)

        if size is not None:
            # NEW: Check for second-order array notation [outer, inner]
            if isinstance(size, list) and len(size) == 2:
                results = _handle_nested_combinations(choices, size)
                # Apply count limit if specified
                if count is not None and len(results) > count:
                    results = random.sample(results, count)
                return results

            # Apply size constraints first, then expand
            vals = []

            # Handle tuple size (from, to) or single size
            if isinstance(size, tuple) and len(size) == 2:
                from_size, to_size = size
                # Generate combinations for all sizes from from_size to to_size (inclusive)
                for s in range(from_size, to_size + 1):
                    if s > len(choices):
                        continue
                    for combo in combinations(choices, s):
                        # Expand this specific combination
                        combo_results = _expand_combination(combo)
                        vals.extend(combo_results)
            else:
                # Single size value
                if size <= len(choices):
                    for combo in combinations(choices, size):
                        # Expand this specific combination
                        combo_results = _expand_combination(combo)
                        vals.extend(combo_results)

            # Apply count limit if specified
            if count is not None and len(vals) > count:
                vals = random.sample(vals, count)

            return vals
        else:
            # Original behavior: expand all choices
            vals = []
            for choice in choices:
                ex = expand_spec(choice)
                # expand_spec returns list; extend with each item (scalar or dict value)
                vals.extend(ex)

            # Apply count limit if specified
            if count is not None and len(vals) > count:
                vals = random.sample(vals, count)

            return vals
    elif isinstance(v, Mapping) and ("_range_" in v.keys()):
        # Handle _range_ in value position
        range_spec = v["_range_"]
        count = v.get("count", None)

        # Generate numeric range based on specification
        range_values = _generate_range(range_spec)

        # Apply count limit if specified
        if count is not None and len(range_values) > count:
            range_values = random.sample(range_values, count)

        return range_values
    elif isinstance(v, Mapping):
        # Nested object: expand to list of dict values
        return expand_spec(v)
    elif isinstance(v, list):
        # Handle lists in value positions
        return expand_spec(v)
    else:
        return [v]

def _count_value(v):
    """Count combinations for a value position."""
    if isinstance(v, Mapping) and ("_or_" in v.keys()):
        return count_combinations(v)
    elif isinstance(v, Mapping) and ("_range_" in v.keys()):
        return count_combinations(v)
    elif isinstance(v, Mapping):
        return count_combinations(v)
    elif isinstance(v, list):
        return count_combinations(v)
    else:
        return 1


def _generate_range(range_spec):
    """Generate numeric range from specification.

    Supports two syntaxes:
    - Array: [from, to, step] where step defaults to 1
    - Dict: {"from": start, "to": end, "step": step}
    """
    if isinstance(range_spec, list):
        if len(range_spec) == 2:
            start, end = range_spec
            step = 1
        elif len(range_spec) == 3:
            start, end, step = range_spec
        else:
            raise ValueError("Range array must be [from, to] or [from, to, step]")
    elif isinstance(range_spec, dict):
        start = range_spec["from"]
        end = range_spec["to"]
        step = range_spec.get("step", 1)
    else:
        raise ValueError("Range specification must be array [from, to, step] or dict {'from': start, 'to': end, 'step': step}")

    # Generate range - end is inclusive
    if step > 0:
        return list(range(start, end + 1, step))
    else:
        # For negative steps, we need to ensure end is included
        return list(range(start, end - 1, step))


def _count_range(range_spec):
    """Count elements in a numeric range without generating them."""
    if isinstance(range_spec, list):
        if len(range_spec) == 2:
            start, end = range_spec
            step = 1
        elif len(range_spec) == 3:
            start, end, step = range_spec
        else:
            raise ValueError("Range array must be [from, to] or [from, to, step]")
    elif isinstance(range_spec, dict):
        start = range_spec["from"]
        end = range_spec["to"]
        step = range_spec.get("step", 1)
    else:
        raise ValueError("Range specification must be array [from, to, step] or dict {'from': start, 'to': end, 'step': step}")

    # Calculate count: (end - start) // step + 1 (end is inclusive)
    if step > 0 and end >= start:
        return (end - start) // step + 1
    elif step < 0 and end <= start:
        return (start - end) // abs(step) + 1
    else:
        return 0



"""
Predictions management using Polars.

This module contains Predictions class for storing and managing model predictions
with metadata using Polars DataFrame as the backend for efficient data manipulation.
"""

import pandas as pd
from typing import Dict, Any, List, Optional, Tuple, Union
import numpy as np
import polars as pl
from pathlib import Path
import json
import hashlib
import nirs4all.dataset.evaluator as Evaluator

class Predictions:
    """
    Storage for model predictions using Polars DataFrame backend.

    Each prediction is stored as a row with the following schema:
    - dataset_name: str
    - dataset_path: str
    - config_name: str
    - config_path: str
    - step_idx: int
    - op_counter: int
    - model_name: str
    - model_classname: str
    - model_path: str
    - fold_id: str
    - sample_indices: List[int] (stored as string)
    - weights: Optional[List[float]] (stored as string)
    - metadata: Dict[str, Any] (stored as JSON string)
    - partition: str
    - y_true: List[float] (stored as string)
    - y_pred: List[float] (stored as string)
    - val_score: Optional[float]
    - test_score: Optional[float]
    - metric: str
    - task_type: str
    - n_samples: int
    - n_features: int
    """

    def __init__(self, filepath: Optional[str] = None):
        """Initialize Predictions storage with Polars DataFrame backend."""
        self._df = pl.DataFrame(schema={
            "id": pl.Utf8,
            "dataset_name": pl.Utf8,
            "dataset_path": pl.Utf8,
            "config_name": pl.Utf8,
            "config_path": pl.Utf8,
            "step_idx": pl.Int64,
            "op_counter": pl.Int64,
            "model_name": pl.Utf8,
            "model_classname": pl.Utf8,
            "model_path": pl.Utf8,
            "fold_id": pl.Utf8,
            "sample_indices": pl.Utf8,  # JSON string
            "weights": pl.Utf8,  # JSON string
            "metadata": pl.Utf8,  # JSON string
            "partition": pl.Utf8,
            "y_true": pl.Utf8,  # JSON string
            "y_pred": pl.Utf8,  # JSON string
            "val_score": pl.Float64,
            "test_score": pl.Float64,
            "metric": pl.Utf8,
            "task_type": pl.Utf8,
            "n_samples": pl.Int64,
            "n_features": pl.Int64,
            "preprocessings": pl.Utf8,
            "best_params": pl.Utf8,  # JSON string
        })

        if filepath and Path(filepath).exists():
            self.load_from_file(filepath)

    @staticmethod
    def _generate_hash(row_dict: Dict[str, Any]) -> str:
        """
        Generate a 6-character hash from row dictionary.
        Uses key identifying fields to create a diverse, reproducible hash.
        """
        # Select key fields that uniquely identify a prediction
        key_fields = [
            str(row_dict.get('dataset_name', '')),
            str(row_dict.get('config_name', '')),
            str(row_dict.get('model_name', '')),
            str(row_dict.get('fold_id', '')),
            str(row_dict.get('step_idx', 0)),
            str(row_dict.get('op_counter', 0)),
        ]

        # Create a string to hash
        hash_string = '|'.join(key_fields)

        # Generate SHA256 hash and take first 6 characters (alphanumeric)
        hash_obj = hashlib.sha256(hash_string.encode('utf-8'))
        hex_hash = hash_obj.hexdigest()

        # Convert to base36 (0-9, a-z) for better diversity in 6 chars
        # Take multiple segments of the hex and combine them
        hash_int = int(hex_hash[:16], 16)  # Use first 16 hex chars

        # Convert to base36 and take 6 characters
        base36_chars = '0123456789abcdefghijklmnopqrstuvwxyz'
        result = ''
        temp = hash_int

        for _ in range(6):
            result = base36_chars[temp % 36] + result
            temp //= 36

        return result

    @staticmethod
    def save_predictions_to_csv(
        y_true: Optional[Union[np.ndarray, List[float]]] = None,
        y_pred: Optional[Union[np.ndarray, List[float]]] = None,
        filepath: str = "",
        prefix: str = "",
        suffix: str = ""
    ) -> None:
        """
        Save y_true and y_pred arrays to a CSV file.

        Args:
            y_true: True values array (optional, can be None for prediction-only mode)
            y_pred: Predicted values array (required)
            filepath: Output CSV file path
            prefix: Optional prefix for column names
            suffix: Optional suffix for column names
        """
        if y_pred is None:
            raise ValueError("y_pred is required")

        # Convert to numpy arrays if needed and flatten
        y_pred_arr = np.array(y_pred) if not isinstance(y_pred, np.ndarray) else y_pred
        y_pred_flat = y_pred_arr.flatten()

        data_dict = {}
        pred_col = f"{prefix}y_pred{suffix}"
        data_dict[pred_col] = y_pred_flat.tolist()

        # Handle y_true if provided
        if y_true is not None:
            y_true_arr = np.array(y_true) if not isinstance(y_true, np.ndarray) else y_true
            y_true_flat = y_true_arr.flatten()

            if len(y_true_flat) != len(y_pred_flat):
                raise ValueError(f"Length mismatch after flattening: y_true ({len(y_true_flat)}) != y_pred ({len(y_pred_flat)})")

            true_col = f"{prefix}y_true{suffix}"
            data_dict[true_col] = y_true_flat.tolist()

        # Create DataFrame and save to CSV
        import polars as pl
        df_csv = pl.DataFrame(data_dict)        # Create directory if it doesn't exist
        Path(filepath).parent.mkdir(parents=True, exist_ok=True)

        # Save to CSV
        df_csv.write_csv(filepath)
        print(f"💾 Saved predictions to {filepath}")

    def add_prediction(
        self,
        dataset_name: str,
        dataset_path: str = "",
        config_name: str = "",
        config_path: str = "",
        step_idx: int = 0,
        op_counter: int = 0,
        model_name: str = "",
        model_classname: str = "",
        model_path: str = "",
        fold_id: Optional[Union[str, int]] = None,
        sample_indices: Optional[List[int]] = None,
        weights: Optional[List[float]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        partition: str = "",
        y_true: Optional[np.ndarray] = None,
        y_pred: Optional[np.ndarray] = None,
        val_score: Optional[float] = None,
        test_score: Optional[float] = None,
        metric: str = "mse",
        task_type: str = "regression",
        n_samples: int = 0,
        n_features: int = 0,
        preprocessings: str = "",
        best_params: Optional[Dict[str, Any]] = None
    ) -> str:
        """Add a new prediction to the storage."""

        # Convert numpy arrays to lists for JSON serialization
        y_true_list = y_true.tolist() if y_true is not None else []
        y_pred_list = y_pred.tolist() if y_pred is not None else []
        sample_indices_list = sample_indices if sample_indices is not None else []
        weights_list = weights if weights is not None else []
        metadata_dict = metadata if metadata is not None else {}
        best_params_dict = best_params if best_params is not None else {}
        fold_id = str(fold_id)

        # Create row_dict with columns in schema order
        row_dict = {
            "id": "",  # Will be filled by hash generation
            "dataset_name": dataset_name,
            "dataset_path": dataset_path,
            "config_name": config_name,
            "config_path": config_path,
            "step_idx": step_idx,
            "op_counter": op_counter,
            "model_name": model_name,
            "model_classname": model_classname,
            "model_path": model_path,
            "fold_id": fold_id,
            "sample_indices": json.dumps(sample_indices_list),
            "weights": json.dumps([w.tolist() if isinstance(w, np.ndarray) else w for w in weights_list]),
            "metadata": json.dumps(metadata_dict),
            "partition": partition,
            "y_true": json.dumps(y_true_list),
            "y_pred": json.dumps(y_pred_list),
            "val_score": val_score,
            "test_score": test_score,
            "metric": metric,
            "task_type": task_type,
            "n_samples": n_samples,
            "n_features": n_features,
            "preprocessings": preprocessings,
            "best_params": json.dumps(best_params_dict),
        }

        # Generate unique ID hash for the prediction
        prediction_id = self._generate_hash(row_dict)
        row_dict["id"] = prediction_id
        new_row = pl.DataFrame([row_dict])
        self._df = pl.concat([self._df, new_row])

        return row_dict["id"]

    def add_predictions(
        self,
        dataset_name: Union[str, List[str]],
        dataset_path: Union[str, List[str]] = "",
        config_name: Union[str, List[str]] = "",
        config_path: Union[str, List[str]] = "",
        step_idx: Union[int, List[int]] = 0,
        op_counter: Union[int, List[int]] = 0,
        model_name: Union[str, List[str]] = "",
        model_classname: Union[str, List[str]] = "",
        model_path: Union[str, List[str]] = "",
        fold_id: Union[Optional[str], List[Optional[str]]] = None,
        sample_indices: Union[Optional[List[int]], List[Optional[List[int]]]] = None,
        weights: Union[Optional[List[float]], List[Optional[List[float]]]] = None,
        metadata: Union[Optional[Dict[str, Any]], List[Optional[Dict[str, Any]]]] = None,
        partition: Union[str, List[str]] = "",
        y_true: Union[Optional[np.ndarray], List[Optional[np.ndarray]]] = None,
        y_pred: Union[Optional[np.ndarray], List[Optional[np.ndarray]]] = None,
        val_score: Union[Optional[float], List[Optional[float]]] = None,
        test_score: Union[Optional[float], List[Optional[float]]] = None,
        metric: Union[str, List[str]] = "mse",
        task_type: Union[str, List[str]] = "regression",
        n_samples: Union[int, List[int]] = 0,
        n_features: Union[int, List[int]] = 0,
        preprocessings: Union[str, List[str]] = "",
        best_params: Union[Optional[Dict[str, Any]], List[Optional[Dict[str, Any]]]] = None
    ) -> None:
        """
        Add multiple predictions to the storage.

        For each parameter:
        - If it's a single value, it will be copied to all predictions
        - If it's a list, the value at each index will be used for the corresponding prediction

        The number of predictions is determined by the longest list parameter.

        Args:
            dataset_name: Dataset name(s) - can be single string or list
            dataset_path: Dataset path(s) - can be single string or list
            config_name: Config name(s) - can be single string or list
            config_path: Config path(s) - can be single string or list
            step_idx: Step index(es) - can be single int or list
            op_counter: Operation counter(s) - can be single int or list
            model_name: Model name(s) - can be single string or list
            model_classname: Model classname(s) - can be single string or list
            model_path: Model path(s) - can be single string or list
            fold_id: Fold ID(s) - can be single string or list
            sample_indices: Sample indices - can be single list or list of lists
            weights: Weights - can be single list or list of lists
            metadata: Metadata - can be single dict or list of dicts
            partition: Partition(s) - can be single string or list
            y_true: True values - can be single array or list of arrays
            y_pred: Predicted values - can be single array or list of arrays
            val_score: Loss score(s) - can be single float or list
            test_score: Evaluation score(s) - can be single float or list
            metric: Metric(s) - can be single string or list
            task_type: Task type(s) - can be single string or list
            n_samples: Number of samples - can be single int or list
            n_features: Number of features - can be single int or list
        """
        # Collect all parameters
        params = {
            'dataset_name': dataset_name,
            'dataset_path': dataset_path,
            'config_name': config_name,
            'config_path': config_path,
            'step_idx': step_idx,
            'op_counter': op_counter,
            'model_name': model_name,
            'model_classname': model_classname,
            'model_path': model_path,
            'fold_id': fold_id,
            'sample_indices': sample_indices,
            'weights': weights,
            'metadata': metadata,
            'partition': partition,
            'y_true': y_true,
            'y_pred': y_pred,
            'val_score': val_score,
            'test_score': test_score,
            'metric': metric,
            'task_type': task_type,
            'n_samples': n_samples,
            'n_features': n_features,
            'preprocessings': preprocessings,
            'best_params': best_params
        }

        # Find the maximum length (number of predictions to create)
        max_length = 1
        list_params = {}

        for param_name, param_value in params.items():
            if isinstance(param_value, list):
                max_length = max(max_length, len(param_value))
                list_params[param_name] = param_value

        if max_length == 1:
            # No lists found, just call add_prediction once
            self.add_prediction(**params)
            return

        # Create individual predictions
        for i in range(max_length):
            prediction_params = {}

            for param_name, param_value in params.items():
                if isinstance(param_value, list):
                    # Use value at index i, or last value if list is shorter
                    idx = min(i, len(param_value) - 1)
                    prediction_params[param_name] = param_value[idx]
                else:
                    # Single value, copy to all predictions
                    prediction_params[param_name] = param_value

            # Add the individual prediction
            self.add_prediction(**prediction_params)

    def filter_predictions(
        self,
        dataset_name: Optional[str] = None,
        partition: Optional[str] = None,
        config_name: Optional[str] = None,
        model_name: Optional[str] = None,
        fold_id: Optional[str] = None,
        step_idx: Optional[int] = None,
        **kwargs
    ) -> List[Dict[str, Any]]:
        """
        Filter predictions and return as list of dictionaries.

        Args:
            dataset_name: Filter by dataset name
            partition: Filter by partition
            config_name: Filter by config name
            model_name: Filter by model name
            fold_id: Filter by fold ID
            step_idx: Filter by step index
            **kwargs: Additional filter criteria

        Returns:
            List of prediction dictionaries
        """
        df_filtered = self._df

        # Apply filters
        if dataset_name is not None:
            df_filtered = df_filtered.filter(pl.col("dataset_name") == dataset_name)
        if partition is not None:
            df_filtered = df_filtered.filter(pl.col("partition") == partition)
        if config_name is not None:
            df_filtered = df_filtered.filter(pl.col("config_name") == config_name)
        if model_name is not None:
            df_filtered = df_filtered.filter(pl.col("model_name") == model_name)
        if fold_id is not None:
            df_filtered = df_filtered.filter(pl.col("fold_id") == fold_id)
        if step_idx is not None:
            df_filtered = df_filtered.filter(pl.col("step_idx") == step_idx)

        # Apply additional filters from kwargs
        for key, value in kwargs.items():
            if key in self._df.columns:
                df_filtered = df_filtered.filter(pl.col(key) == value)

        # Convert to list of dictionaries with JSON deserialization
        results = []
        for row in df_filtered.to_dicts():
            # Deserialize JSON fields
            row["sample_indices"] = json.loads(row["sample_indices"])
            row["weights"] = json.loads(row["weights"])
            row["metadata"] = json.loads(row["metadata"])
            row["best_params"] = json.loads(row["best_params"]) if row["best_params"] else {}
            row["y_true"] = np.array(json.loads(row["y_true"]))
            row["y_pred"] = np.array(json.loads(row["y_pred"]))
            results.append(row)

        return results

    def get_similar(self, **filter_kwargs) -> Optional[Dict[str, Any]]:
        """
        Get the first prediction similar to the provided filter criteria.

        Args:
            **filter_kwargs: Filter criteria (same as filter_predictions)

        Returns:
            First matching prediction as dictionary, or None if no matches found
        """
        results = self.filter_predictions(**filter_kwargs)
        return results[0] if results else None

    @property
    def num_predictions(self) -> int:
        """Get the number of stored predictions."""
        return len(self._df)

    def get_unique_values(self, column: str) -> List[str]:
        """Get unique values for a specific column."""
        if column not in self._df.columns:
            raise ValueError(f"Column '{column}' not found in predictions")
        return self._df[column].unique().to_list()

    def get_datasets(self) -> List[str]:
        """Get list of unique dataset names."""
        return self.get_unique_values("dataset_name")

    def get_partitions(self) -> List[str]:
        """Get list of unique partitions."""
        return self.get_unique_values("partition")

    def get_configs(self) -> List[str]:
        """Get list of unique config names."""
        return self.get_unique_values("config_name")

    def get_models(self) -> List[str]:
        """Get list of unique model names."""
        return self.get_unique_values("model_name")

    def list_keys(self) -> List[str]:
        """Get list of unique prediction keys (for compatibility with old interface)."""
        # Generate keys similar to the old format: dataset/config/model/partition/fold
        if len(self._df) == 0:
            return []

        keys = []
        for row in self._df.iter_rows(named=True):
            fold_part = f"_fold_{row['fold_id']}" if row['fold_id'] is not None else ""
            key = f"{row['dataset_name']}/{row['config_name']}/{row['model_name']}/{row['partition']}{fold_part}"
            keys.append(key)

        return list(set(keys))  # Remove duplicates

    # def get_prediction_by_key(self, key: str) -> Optional[Dict[str, Any]]:
    #     """Get prediction data by key (for compatibility with old interface)."""
    #     # Parse key format: dataset/config/model/partition[_fold_X]
    #     parts = key.split('/')
    #     if len(parts) < 4:
    #         return None

    #     dataset_name, config_name, model_name, partition_part = parts[:4]

    #     # Extract fold info if present
    #     fold_id = None
    #     partition = partition_part
    #     if '_fold_' in partition_part:
    #         partition, fold_part = partition_part.split('_fold_')
    #         try:
    #             fold_id = int(fold_part)
    #         except ValueError:
    #             pass

    #     # Query the DataFrame
    #     filter_expr = (
    #         (pl.col('dataset_name') == dataset_name) &
    #         (pl.col('config_name') == config_name) &
    #         (pl.col('model_name') == model_name) &
    #         (pl.col('partition') == partition)
    #     )

    #     if fold_id is not None:
    #         filter_expr = filter_expr & (pl.col('fold_id') == fold_id)

    #     matches = self._df.filter(filter_expr)

    #     if len(matches) == 0:
    #         return None

    #     # Return the first match as a dictionary with the expected format
    #     row = matches.row(0, named=True)
    #     return {
    #         'dataset_name': row['dataset_name'],
    #         'config_name': row['config_name'],
    #         'model_name': row['model_name'],
    #         'partition': row['partition'],
    #         'fold_id': row['fold_id'],
    #         'y_true': json.loads(row['y_true']) if row['y_true'] else [],
    #         'y_pred': json.loads(row['y_pred']) if row['y_pred'] else [],
    #         'test_score': row['test_score'],
    #         'metric': row['metric'],
    #     }

    def save_to_file(self, filepath: str) -> None:
        """Save predictions to JSON file."""
        try:
            Path(filepath).parent.mkdir(parents=True, exist_ok=True)

            # Convert DataFrame to JSON-serializable format
            data = self._df.to_dicts()

            with open(filepath, 'w') as f:
                json.dump(data, f, indent=2)

            # print(f"💾 Saved {len(self._df)} predictions to {filepath}")

        except Exception as e:
            print(f"⚠️ Error saving predictions to {filepath}: {e}")

    def load_from_file(self, filepath: str) -> None:
        """Load predictions from JSON file."""
        try:
            with open(filepath, 'r') as f:
                data = json.load(f)

            if data:
                self._df = pl.DataFrame(data)
                # print(f"📥 Loaded {len(self._df)} predictions from {filepath}")

        except Exception as e:
            pass
            # print(f"⚠️ Error loading predictions from {filepath}: {e}")

    @classmethod
    def load_from_file_cls(cls, filepath: str) -> 'Predictions':
        """Load predictions from JSON file as class method."""
        instance = cls()
        instance.load_from_file(filepath)
        return instance

    def save_prediction_to_csv(self, filepath: str, index: Optional[int] = None) -> None:
        """
        Save a single prediction to CSV file.

        Args:
            filepath: Output CSV file path
            index: Index of prediction to save (if None, saves all)
        """
        try:
            Path(filepath).parent.mkdir(parents=True, exist_ok=True)

            if index is not None:
                # Save single prediction
                if index >= len(self._df):
                    raise IndexError(f"Index {index} out of range")
                row = self._df[index].to_dicts()[0]

                # Deserialize arrays for CSV
                y_true = json.loads(row["y_true"])
                y_pred = json.loads(row["y_pred"])
                sample_indices = json.loads(row["sample_indices"])

                # Create CSV data
                csv_data = []
                for i, (true_val, pred_val) in enumerate(zip(y_true, y_pred)):
                    sample_idx = sample_indices[i] if i < len(sample_indices) else i
                    csv_data.append({
                        "sample_index": sample_idx,
                        "y_true": true_val,
                        "y_pred": pred_val,
                        "dataset_name": row["dataset_name"],
                        "model_name": row["model_name"],
                        "partition": row["partition"],
                        "fold_id": row["fold_id"],
                        "metric": row["metric"]
                    })

                df_csv = pl.DataFrame(csv_data)
                df_csv.write_csv(filepath)
                print(f"💾 Saved prediction {index} to {filepath}")
            else:
                # Save all predictions in expanded format
                self._df.write_csv(filepath)
                print(f"💾 Saved all {len(self._df)} predictions to {filepath}")

        except Exception as e:
            print(f"⚠️ Error saving prediction to CSV {filepath}: {e}")

    def calculate_scores(self, metrics: Optional[List[str]] = None) -> pl.DataFrame:
        """
        Calculate scores for all predictions using external evaluator.

        Args:
            metrics: List of metrics to calculate (defaults to ['mse', 'r2', 'mae'])

        Returns:
            DataFrame with calculated scores
        """
        if metrics is None:
            metrics = ['mse', 'r2', 'mae']

        # Import the external evaluator
        try:
            from ..utils.model_utils import ModelUtils, TaskType
            model_utils = ModelUtils()
        except ImportError:
            print("⚠️ Cannot import ModelUtils for score calculation")
            return pl.DataFrame()

        scores_data = []

        for row in self._df.to_dicts():
            # Deserialize prediction arrays
            y_true = np.array(json.loads(row["y_true"]))
            y_pred = np.array(json.loads(row["y_pred"]))

            if len(y_true) > 0 and len(y_pred) > 0:
                # Calculate all scores using regression as default
                all_scores = model_utils.calculate_scores(y_true, y_pred, TaskType.REGRESSION)

                # Filter to requested metrics if they exist
                scores = {}
                for metric in metrics:
                    if metric in all_scores:
                        scores[metric] = all_scores[metric]
                    else:
                        # If metric not found, set to None
                        scores[metric] = None

                # Create score record
                score_record = {
                    "dataset_name": row["dataset_name"],
                    "config_name": row["config_name"],
                    "model_name": row["model_name"],
                    "partition": row["partition"],
                    "fold_id": row["fold_id"],
                    "step_idx": row["step_idx"],
                }

                # Add calculated metrics
                for metric, score in scores.items():
                    score_record[f"score_{metric}"] = score if isinstance(score, (int, float)) else None

                scores_data.append(score_record)

        return pl.DataFrame(scores_data)

    def top_k(self, k: int = 5, metric: str = "", ascending: bool = True, aggregate_partitions: List[str] = [], **filters) -> List[Dict[str, Any]]:
        """
        Get top K predictions ranked by metric, val_score, or test_score.
        By default filters to test partition unless otherwise specified.

        Args:
            metric: Metric name to rank by ("" for test_score, "loss" for val_score, else calculate metric on-the-fly)
            k: Number of top results to return (-1 to return all filtered predictions)
            ascending: If True, lower scores rank higher (for error metrics)
            aggregate_partitions: List of partitions to aggregate y_true and y_pred from (e.g. ['train', 'val'])
            **filters: Additional filter criteria

        Returns:
            List of top K prediction dictionaries (or all if k=-1)
        """
        # Add default partition filter if not specified
        if 'partition' not in filters:
            filters['partition'] = 'val'
        if 'partition' in filters and filters['partition'] in ['all', 'ALL', 'All', '_all_', '']:
            del filters['partition']

        # First filter the entries
        df_filtered = self._df
        for key, value in filters.items():
            if key in df_filtered.columns:
                df_filtered = df_filtered.filter(pl.col(key) == value)

        # print( f"🔍 Found {len(df_filtered)} predictions after filtering with criteria: {filters}")

        if df_filtered.is_empty():
            return []

        # Handle different ranking scenarios
        if metric == "" or metric == "loss":
            # Use existing stored scores
            rank_col = "val_score"
            df_ranked = df_filtered.filter(pl.col(rank_col).is_not_null())

            if df_ranked.is_empty():
                return []

            df_sorted = df_ranked.sort(rank_col, descending=not ascending)

            # Return all results if k=-1, otherwise return top k
            if k == -1:
                top_k_rows = df_sorted
            else:
                top_k_rows = df_sorted.head(k)

            # Convert to list of dictionaries with JSON deserialization
            results = []
            for row in top_k_rows.to_dicts():
                # Deserialize JSON fields
                row["sample_indices"] = json.loads(row["sample_indices"])
                row["weights"] = json.loads(row["weights"]) if row["weights"] else []
                row["metadata"] = json.loads(row["metadata"]) if row["metadata"] else {}
                row["best_params"] = json.loads(row["best_params"]) if row["best_params"] else {}
                row["y_true"] = np.array(json.loads(row["y_true"]))
                row["y_pred"] = np.array(json.loads(row["y_pred"]))
                results.append(row)

            # Add partition data if requested
            if len(aggregate_partitions) > 0:
                results = self._add_partition_data(results, aggregate_partitions)
            return results

        else:
            # Calculate metric on-the-fly for all filtered entries
            scores_data = []

            # Import the external evaluator
            try:
                from ..utils.model_utils import ModelUtils, TaskType
                model_utils = ModelUtils()
            except ImportError:
                print("⚠️ Cannot import ModelUtils for score calculation")
                return []

            for i, row in enumerate(df_filtered.to_dicts()):
                # Deserialize prediction arrays
                y_true = np.array(json.loads(row["y_true"]))
                y_pred = np.array(json.loads(row["y_pred"]))

                if len(y_true) > 0 and len(y_pred) > 0:
                    # Calculate all scores using regression as default, explicitly include the requested metric
                    all_scores = model_utils.calculate_scores(y_true, y_pred, TaskType.REGRESSION, metrics=[metric])

                    # Get the requested metric score
                    if metric in all_scores:
                        metric_score = all_scores[metric]

                        # Create score record
                        score_record = row.copy()
                        score_record["computed_score"] = metric_score
                        score_record["computed_metric"] = metric
                        score_record[metric] = metric_score

                        # Deserialize JSON fields for final output
                        score_record["sample_indices"] = json.loads(row["sample_indices"])
                        score_record["weights"] = json.loads(row["weights"]) if row["weights"] else []
                        score_record["metadata"] = json.loads(row["metadata"]) if row["metadata"] else {}
                        score_record["best_params"] = json.loads(row["best_params"]) if row["best_params"] else {}
                        score_record["y_true"] = y_true
                        score_record["y_pred"] = y_pred

                        scores_data.append(score_record)

            # Sort by computed metric and return top k (or all if k=-1)
            if not scores_data:
                return []
            if ModelUtils._is_higher_better(metric):
                ascending = not ascending  # Reverse for higher is better
            scores_data.sort(key=lambda x: x["computed_score"], reverse=not ascending)

            # Return all results if k=-1, otherwise return top k
            if k == -1:
                results = scores_data
            else:
                results = scores_data[:k]
            # Add partition data if requested
            if len(aggregate_partitions) > 0:
                results = self._add_partition_data(results, aggregate_partitions)

            return results

    def _add_partition_data(self, results: List[Dict[str, Any]], aggregate_partitions: List[str]) -> List[Dict[str, Any]]:
        """
        Add y_true and y_pred data from all partitions to each result using simple filtering.

        Args:
            results: List of prediction results

        Returns:
            Results with added partition structure: train: {y_true: ..., y_pred: ...}, val: {...}, test: {...}
        """
        for result in results:
            # For each partition, filter once and get the data
            for partition in aggregate_partitions:
                partition_data = self.filter_predictions(
                    dataset_name=result['dataset_name'],
                    config_name=result['config_name'],
                    model_name=result['model_name'],
                    fold_id=result['fold_id'],
                    step_idx=result['step_idx'],
                    op_counter=result['op_counter'],
                    partition=partition
                )

                if partition_data:
                    # Found data for this partition
                    result[partition] = {
                        'y_true': partition_data[0]['y_true'],
                        'y_pred': partition_data[0]['y_pred']
                    }
                else:
                    # No data for this partition
                    result[partition] = {
                        'y_true': np.array([]),
                        'y_pred': np.array([])
                    }

        return results

    def get_best(self, metric: str = "", ascending: bool = True, aggregate_partitions: List[str] = [], **filters) -> Optional[Dict[str, Any]]:
        """
        Get the best prediction for a specific metric, val_score, or test_score.
        This is an alias for top_k with k=1.

        Args:
            metric: Metric name to optimize ("" for test_score, "loss" for val_score, else metric)
            ascending: If True, lower scores are better (for error metrics)
            aggregate_partitions: If True, add y_true and y_pred for all partitions (train, val, test)
            **filters: Additional filter criteria

        Returns:
            Best prediction dictionary or None
        """
        top_results = self.top_k(k=1, metric=metric, ascending=ascending, aggregate_partitions=aggregate_partitions, **filters)
        return top_results[0] if top_results else None

    def bottom_k(self, metric: str = "", k: int = 5, aggregate_partitions: List[str] = [], **filters) -> List[Dict[str, Any]]:
        """
        Get bottom K predictions (worst performing).
        This is an alias for top_k with ascending=False.
        By default filters to test partition unless otherwise specified.

        Args:
            metric: Metric name to rank by ("" for test_score, "loss" for val_score, else metric)
            k: Number of bottom results to return (-1 to return all filtered predictions)
            aggregate_partitions: List of partitions to aggregate y_true and y_pred from (e.g. ['train', 'val'])
            **filters: Additional filter criteria

        Returns:
            List of bottom K prediction dictionaries (or all if k=-1)
        """
        return self.top_k(k=k, metric=metric, ascending=False, aggregate_partitions=aggregate_partitions, **filters)

    def clear(self) -> None:
        """Clear all predictions."""
        self._df = self._df.clear()

    def merge_predictions(self, other: 'Predictions') -> None:
        """
        Merge predictions from another Predictions instance.

        Args:
            other: Another Predictions instance to merge from

        Note:
            - Duplicate predictions (same metadata) will be kept (no deduplication)
            - Use this method to combine results from multiple experiments
        """
        if not isinstance(other, Predictions):
            raise TypeError("Can only merge with another Predictions instance")

        if len(other._df) == 0:
            print("⚠️ No predictions to merge (source is empty)")
            return

        # Ensure schemas are compatible before concatenating
        if len(self._df) == 0:
            # If current DataFrame is empty, just copy the other
            self._df = other._df.clone()
        else:
            # Check if schemas are compatible and align them before concatenating
            self_schema = self._df.schema
            other_schema = other._df.schema

            # Check if schemas match exactly
            schemas_match = (
                (len(self_schema) == len(other_schema))
                and all(col in other_schema and self_schema[col] == other_schema[col] for col in self_schema)
                and (list(self._df.columns) == list(other._df.columns))
            )

            if schemas_match:
                # Schemas match, safe to concatenate directly
                self._df = pl.concat([self._df, other._df], how="vertical")
            else:
                # Schemas don't match, need to align them
                # print(f"⚠️ Schema mismatch detected, aligning schemas before merge")

                # Use the predefined schema order from __init__ to ensure consistency
                predefined_order = [
                    "id", "dataset_name", "dataset_path", "config_name", "config_path",
                    "step_idx", "op_counter", "model_name", "model_classname", "model_path",
                    "fold_id", "sample_indices", "weights", "metadata", "partition",
                    "y_true", "y_pred", "val_score", "test_score", "metric", "task_type",
                    "n_samples", "n_features", "preprocessings", "best_params"
                ]

                # Determine the target schema by preferring non-null types, maintaining order
                unified_schema = {}
                all_columns = set(self_schema.keys()) | set(other_schema.keys())

                # Process columns in predefined order first, then any extra columns
                ordered_columns = [col for col in predefined_order if col in all_columns]
                extra_columns = [col for col in all_columns if col not in predefined_order]
                ordered_columns.extend(extra_columns)

                for col_name in ordered_columns:
                    self_type = self_schema.get(col_name, pl.Null)
                    other_type = other_schema.get(col_name, pl.Null)

                    # Prefer non-null types
                    if self_type == pl.Null and other_type != pl.Null:
                        unified_schema[col_name] = other_type
                    elif other_type == pl.Null and self_type != pl.Null:
                        unified_schema[col_name] = self_type
                    elif self_type == other_type:
                        unified_schema[col_name] = self_type
                    else:
                        # If types differ and neither is null, prefer the more specific type
                        # Float64 > Int64 > Utf8 > Null in terms of preference
                        type_priority = {pl.Null: 0, pl.Utf8: 1, pl.Int64: 2, pl.Float64: 3}
                        self_priority = type_priority.get(self_type, 1)
                        other_priority = type_priority.get(other_type, 1)

                        if other_priority > self_priority:
                            unified_schema[col_name] = other_type
                        else:
                            unified_schema[col_name] = self_type

                # Align self DataFrame to unified schema (maintaining column order)
                self_cast_expressions = []
                for col_name in ordered_columns:
                    target_type = unified_schema[col_name]
                    if col_name in self._df.columns:
                        if self._df[col_name].dtype == pl.Null:
                            self_cast_expressions.append(pl.lit(None).cast(target_type).alias(col_name))
                        else:
                            self_cast_expressions.append(pl.col(col_name).cast(target_type))
                    else:
                        self_cast_expressions.append(pl.lit(None).cast(target_type).alias(col_name))

                # Align other DataFrame to unified schema (maintaining column order)
                other_cast_expressions = []
                for col_name in ordered_columns:
                    target_type = unified_schema[col_name]
                    if col_name in other._df.columns:
                        if other._df[col_name].dtype == pl.Null:
                            other_cast_expressions.append(pl.lit(None).cast(target_type).alias(col_name))
                        else:
                            other_cast_expressions.append(pl.col(col_name).cast(target_type))
                    else:
                        other_cast_expressions.append(pl.lit(None).cast(target_type).alias(col_name))

                # Apply schema alignment
                self_aligned = self._df.select(self_cast_expressions)
                other_aligned = other._df.select(other_cast_expressions)

                # Concatenate aligned DataFrames
                # Concatenate aligned DataFrames
                self._df = pl.concat([self_aligned, other_aligned], how="vertical")

    def merge_predictions_with_dedup(self, other: 'Predictions') -> None:
        """
        Merge predictions from another Predictions instance with deduplication.

        Args:
            other: Another Predictions instance to merge from

        Note:
            - Duplicates are identified by: dataset_name, config_name, model_name,
              partition, fold_id, step_idx, op_counter
            - When duplicates are found, keeps the existing prediction (no replacement)
        """
        if not isinstance(other, Predictions):
            raise TypeError("Can only merge with another Predictions instance")

        if len(other._df) == 0:
            print("⚠️ No predictions to merge (source is empty)")
            return

        original_count = len(self._df)

        # Define key columns for duplicate detection (excluding 'id' since it's auto-generated)
        key_columns = [
            "dataset_name", "config_name", "model_name",
            "partition", "fold_id", "step_idx", "op_counter"
        ]

        # Create a combined DataFrame
        combined_df = pl.concat([self._df, other._df], how="vertical")

        # Remove duplicates keeping the first occurrence (existing predictions)
        deduplicated_df = combined_df.unique(subset=key_columns, keep="first")

        self._df = deduplicated_df

        added_count = len(self._df) - original_count
        duplicates_count = len(other._df) - added_count

        # print(f"✅ Merged {added_count} new predictions, skipped {duplicates_count} duplicates. Total: {len(self._df)} predictions")

    def __len__(self) -> int:
        """Return number of stored predictions."""
        return len(self._df)

    def __repr__(self) -> str:
        if len(self._df) == 0:
            return "Predictions(empty)"
        return f"Predictions({len(self._df)} entries)"

    def __str__(self) -> str:
        if len(self._df) == 0:
            return "📈 Predictions: No predictions stored"

        datasets = self.get_datasets()
        configs = self.get_configs()
        models = self.get_models()

        return (f"📈 Predictions: {len(self._df)} entries\n"
                f"   Datasets: {datasets}\n"
                f"   Configs: {configs}\n"
                f"   Models: {models}")

    def get_entry_partitions(self, entry):
        res = {}
        filter = {
            'dataset_name': entry['dataset_name'],
            'config_name': entry['config_name'],
            'model_name': entry['model_name'],
            'fold_id': entry['fold_id'],
            'step_idx': entry['step_idx'],
            'op_counter': entry['op_counter']
        }

        for partition in ['train', 'val', 'test']:
            filter['partition'] = partition
            predictions = self.filter_predictions(**filter)
            if not predictions or len(predictions) == 0:
                print(f"⚠️ No predictions found for {filter}")
                res[partition] = None
            else:
                res[partition] = predictions[0]
        return res

    @classmethod
    def pred_short_string(cls, entry, metrics=None):
        scores_str = ""
        metrics.remove('rmse') if metrics is not None and 'rmse' in metrics else None
        if metrics is not None:
            scores = Evaluator.eval_list(entry['y_true'], entry['y_pred'], metrics=metrics)
            scores_str = ", ".join([f"[{k}:{v:.4f}]" if k != 'rmse' else f"[{k}:{v:.4f}]" for k, v in zip(metrics, scores)])

        short_desc = f"{entry['model_name']} - {entry['metric']} [test: {entry['test_score']:.4f}], [val: {entry['val_score']:.4f}]"
        short_desc += f", {scores_str}"
        short_desc += f", (fold: {entry['fold_id']}, id: {entry['op_counter']}, step: {entry['step_idx']}) - [{entry['id']}]"
        return short_desc

    @classmethod
    def pred_long_string(cls, entry, metrics=None):  # ADAPT TO CLASSIFICATION
        return Predictions.pred_short_string(entry, metrics=metrics) + f" | [{entry['config_name']}]"

    def to_numpy(self) -> np.ndarray:
        """
        Get all predictions as a numpy array.

        Returns:
            List of all prediction dictionaries
        """
        return self._df.to_numpy()

    def to_list(self) -> List[Dict[str, Any]]:
        """
        Get all predictions as a list of dictionaries.

        Returns:
            List of all prediction dictionaries
        """
        return self.to_numpy().tolist()

    def to_dicts(self) -> List[Dict[str, Any]]:
        """
        Get all predictions as a list of dictionaries.

        Returns:
            List of all prediction dictionaries
        """
        return self._df.to_dicts()

    def to_pandas(self) -> 'pd.DataFrame':
        """
        Get all predictions as a pandas DataFrame.

        Returns:
            Pandas DataFrame of all predictions
        """
        return self._df.to_pandas()




"""
Utilitaires pour détecter les backends ML disponibles et permettre des tests conditionnels.
"""
import importlib.util

TF_AVAILABLE = importlib.util.find_spec('tensorflow') is not None
TORCH_AVAILABLE = importlib.util.find_spec('torch') is not None


def framework(framework_name):
    def decorator(func):
        func.framework = framework_name
        return func
    return decorator


def is_tensorflow_available():
    """Vérifie si TensorFlow est installé."""
    try:
        import tensorflow
        return True
    except ImportError:
        return False


def is_torch_available():
    """Vérifie si PyTorch est installé."""
    try:
        import torch
        return True
    except ImportError:
        return False


def is_keras_available():
    """Vérifie si Keras 3 est installé."""
    try:
        import keras
        return True
    except ImportError:
        return False


def is_jax_available():
    """Vérifie si JAX est installé."""
    try:
        import jax
        return True
    except ImportError:
        return False


def is_gpu_available():
    """
    Vérifie si un GPU est disponible pour au moins un des frameworks installés.
    """
    # Vérifier la disponibilité de GPU pour PyTorch
    if is_torch_available():
        try:
            import torch
            return torch.cuda.is_available()
        except Exception:
            pass

    # Vérifier la disponibilité de GPU pour TensorFlow
    if is_tensorflow_available():
        try:
            import tensorflow as tf
            return len(tf.config.list_physical_devices('GPU')) > 0
        except Exception:
            pass

    # Vérifier la disponibilité de GPU pour JAX
    if is_jax_available():
        try:
            import jax
            return jax.default_backend() == 'gpu'
        except Exception:
            pass

    # Aucun backend GPU disponible
    return False



import csv
import io
import pandas as pd
import gzip
import zipfile
from pathlib import Path
import numpy as np
from typing import Union


# =============================================================================
# Utility: Check if a string can be converted to float, given a decimal separator
# =============================================================================
def _can_be_float(value, decimal_sep):
    """Check if a string can be converted to a float."""
    if not isinstance(value, str):
        return False  # Only strings should be checked

    value = value.strip()
    if not value:
        return False

    try:
        # If there's 'e' or 'E', handle scientific notation
        if 'e' in value.lower() or 'E' in value:
            float(value)
            return True

        # Replace decimal separator if needed
        if decimal_sep == '.':
            float(value)
        else:
            float(value.replace(decimal_sep, '.', 1))
        return True
    except ValueError:
        return False


# =============================================================================
# Utility: Strip all quotes from the file content
# =============================================================================
def _strip_all_quotes(content):
    """
    Removes *all* single-quote and double-quote characters from the string.
    """
    return content.replace('"', '').replace("'", "")


# =============================================================================
# Detect delimiter (unless specified by user)
# =============================================================================
def _detect_delimiter(lines, possible_delimiters=None):
    """
    Attempt to detect the delimiter by looking at the consistency of
    the number of columns. Return the best candidate or None if no good guess.
    """
    if possible_delimiters is None:
        possible_delimiters = [';', ',', '\t', '|', ' ']

    best_delim = None
    max_consistent_cols = -1
    most_cols_at_max_consistency = 0

    # Join lines so csv.reader sees them as input
    content_for_test = "".join(lines)

    for delim_candidate in possible_delimiters:
        try:
            reader = csv.reader(io.StringIO(content_for_test), delimiter=delim_candidate)
            cols_counts = [len(row) for row in reader if row]

            if not cols_counts:
                continue

            # The number of columns that appears the most
            most_frequent_cols = max(set(cols_counts), key=cols_counts.count)
            # How many lines have that number of columns
            consistency = sum(1 for count in cols_counts if count == most_frequent_cols)

            # Choose the delimiter that maximizes the consistency, then the number of columns
            if consistency > max_consistent_cols:
                max_consistent_cols = consistency
                most_cols_at_max_consistency = most_frequent_cols
                best_delim = delim_candidate
            elif consistency == max_consistent_cols:
                # If same consistency, prefer the one with more columns
                if most_frequent_cols > most_cols_at_max_consistency:
                    most_cols_at_max_consistency = most_frequent_cols
                    best_delim = delim_candidate
        except (csv.Error, ValueError):
            continue  # ignore parse errors with this candidate

    return best_delim


# =============================================================================
# Detect decimal separator and header (unless specified by user)
# =============================================================================
def _detect_decimal_and_header(parsed_rows, data_type='x'):
    """
    Given a list of parsed_rows (already split by delimiter),
    try to determine the decimal separator and whether there's a header.

    Returns: (best_decimal_sep, best_has_header)
    """
    if not parsed_rows:
        return '.', False  # fallback

    # We'll guess it by looking at numeric vs. non-numeric content
    num_cols = len(parsed_rows[0])
    if num_cols == 0:
        return '.', False  # fallback

    best_decimal_sep = '.'
    best_has_header = False
    max_numeric_score = -1.0

    for decimal_sep in ['.', ',']:
        for has_header_option in [False, True]:
            first_data_row_index = 1 if has_header_option else 0
            if len(parsed_rows) <= first_data_row_index:
                # no data rows to evaluate
                current_score = 0.0
            else:
                data_rows = parsed_rows[first_data_row_index:]
                numeric_cells = 0
                total_cells = 0

                for row in data_rows:
                    # We only consider rows with at least close to the expected columns
                    if abs(len(row) - num_cols) <= 1:
                        for val in row:
                            total_cells += 1
                            if _can_be_float(val, decimal_sep):  # This would call the commented out function
                                numeric_cells += 1

                current_score = numeric_cells / total_cells if total_cells else 0.0

            # If we declared there's a header but that row also looks numeric,
            # apply a small penalty
            if has_header_option and parsed_rows:
                header_row = parsed_rows[0]
                if len(header_row) == num_cols:
                    header_numeric_cells = sum(_can_be_float(cell, decimal_sep) for cell in header_row)  # This would call the commented out function
                    header_score = header_numeric_cells / len(header_row) if header_row else 0.0
                    if current_score > 0.5 and header_score >= current_score:
                        current_score *= 0.5

            if current_score > max_numeric_score + 1e-6:
                max_numeric_score = current_score
                best_decimal_sep = decimal_sep
                best_has_header = has_header_option
            elif abs(current_score - max_numeric_score) < 1e-6:
                # Tie-break: prefer '.' over ',' and prefer has_header=False over True
                if best_decimal_sep == ',' and decimal_sep == '.':
                    best_decimal_sep = decimal_sep
                    best_has_header = has_header_option
                elif best_has_header and (not has_header_option):
                    best_decimal_sep = decimal_sep
                    best_has_header = has_header_option

    return best_decimal_sep, best_has_header


# =============================================================================
# Main routine: Determine CSV parameters, skipping detection if user param is given
# =============================================================================
def _determine_csv_parameters(csv_content: str,  # csv_content is not used anymore
                              sample_lines=20,  # sample_lines is not used anymore
                              data_type='x',  # data_type is not used anymore
                              user_params=None, *, bypass_auto_detection=True):
    """
    Sets default CSV parameters (delimiter, decimal separator, header)
    and allows them to be overridden by `user_params`.
    The auto-detection logic is commented out.
    """
    if user_params is None:
        user_params = {}

    # Default parameters
    delimiter = user_params.get('delimiter', ';')
    decimal_sep = user_params.get('decimal_separator', '.')
    has_header = user_params.get('has_header', True)

    if not bypass_auto_detection:
        lines = []
        with io.StringIO(csv_content) as f:
            for i, line in enumerate(f):
                if i >= sample_lines:
                    break
                if line.strip():
                    lines.append(line)

        if not lines:
            # no lines to parse
            return {
                'delimiter': user_params.get('delimiter', ';'),  # Default
                'decimal_separator': user_params.get('decimal_separator', '.'),  # Default
                'has_header': user_params.get('has_header', True)  # Default
            }

        # 1) Delimiter detection
        if 'delimiter' in user_params:
            delimiter = user_params['delimiter']
        else:
            delimiter = _detect_delimiter(lines)  # Auto-detection commented out
            # delimiter = ';'  # Default

        if not delimiter:
            delimiter = ';'  # Default

        # 2) Parse a small sample using the chosen delimiter to create parsed_rows
        sample_data = "".join(lines)
        parsed_rows_reader = csv.reader(io.StringIO(sample_data), delimiter=delimiter)
        parsed_rows = [row for row in parsed_rows_reader if any(cell.strip() for cell in row)]

        # 3) Detect decimal separator / header if not specified
        if 'decimal_separator' in user_params:
            decimal_sep = user_params['decimal_separator']
        else:
            decimal_sep, _ = _detect_decimal_and_header(parsed_rows, data_type=data_type)  # Auto-detection commented out
            # decimal_sep = '.'  # Default

        if 'has_header' in user_params:
            has_header = user_params['has_header']
        else:
            _, has_header = _detect_decimal_and_header(parsed_rows, data_type=data_type)  # Auto-detection commented out
            # has_header = True  # Default

    return {
        'delimiter': delimiter,
        'decimal_separator': decimal_sep,
        'has_header': has_header
    }


# =============================================================================
# Main function: load_csv
# =============================================================================
def load_csv(path, na_policy='auto', data_type='x', categorical_mode='auto', **user_params):
    """
    Loads a CSV file using specified or default parameters, cleans data,
    handles NA values, and performs type conversions.

    Args:
        path (str or Path): Path to the CSV file (.csv, .gz, .zip).
        na_policy (str): 'remove' or 'abort' (or 'auto' which acts like 'remove').
            This policy applies to row removal if NAs are found.
        data_type (str): 'x' or 'y'. Influences type conversion.
        categorical_mode (str): How to handle string columns in 'y' data:
            - 'auto': Convert string columns to numerical categories.
            - 'preserve': Keep string columns (will become NaN if not convertible by final astype).
            - 'none': Treat all columns as potentially numeric.
        **user_params: CSV parsing parameters (delimiter, decimal_separator, has_header)
            and other pandas.read_csv arguments.

    Returns:
        (Union[pandas.DataFrame, None], dict, Union[pandas.Series, None]):
            - DataFrame with processed data (before NA row removal).
            - Report dictionary.
            - Boolean Series indicating rows with NAs (aligned with the returned DataFrame).
            None if an error occurs before this stage.
    """
    if na_policy == 'auto':
        na_policy = 'remove'

    if na_policy not in ['remove', 'abort']:
        raise ValueError("Invalid NA policy - only 'remove' or 'abort' (or 'auto') are supported.")
    if categorical_mode not in ['auto', 'preserve', 'none']:
        raise ValueError("Invalid categorical mode - only 'auto', 'preserve', or 'none' are supported.")

    report = {
        'file_path': str(path),
        'detection_params': None,
        'delimiter': None,  # For backward compatibility
        'decimal_separator': None,  # If needed
        'has_header': None,  # If needed
        'initial_shape': None,
        'final_shape': None,
        'na_handling': {
            'strategy': na_policy,
            'na_detected': False,
            'nb_removed_rows': 0,
            'removed_rows_indices': []
        },
        'categorical_info': {},  # Store category mappings
        'warnings': [],  # Store warnings about ambiguous detections
        'error': None
    }

    try:
        file_path = Path(path)
        if not file_path.exists():
            # The test expects either "n'existe pas" or "not exist" in the error
            raise FileNotFoundError(f"Le fichier n'existe pas: {path}")

        # --- 1) Read file content ---
        if file_path.suffix == '.gz':
            with gzip.open(file_path, 'rt', encoding='utf-8') as f:
                content = f.read()
        elif file_path.suffix == '.zip':
            with zipfile.ZipFile(file_path, 'r') as z:
                csv_files_in_zip = [n for n in z.namelist() if n.lower().endswith('.csv')]
                if not csv_files_in_zip:
                    raise ValueError(f"No .csv found in zip: {path}")
                if len(csv_files_in_zip) > 1:
                    print(f"Warning: multiple .csv found in {path}. Using {csv_files_in_zip[0]}")
                content = z.read(csv_files_in_zip[0]).decode('utf-8')
        else:
            # plain text read
            try:
                with open(file_path, 'r', encoding='utf-8', newline='') as f:
                    content = f.read()
            except UnicodeDecodeError:
                print(f"Warning: failed reading {path} with UTF-8. Trying Latin-1.")
                with open(file_path, 'r', encoding='latin-1', newline='') as f:
                    content = f.read()

        if not content.strip():
            raise ValueError("File is empty or could not be read.")

        # --- 2) Determine CSV parameters (now uses defaults or user_params) ---
        # The csv_content argument to _determine_csv_parameters is not strictly needed anymore
        # as auto-detection is off, but kept for signature consistency for now.
        detection_params = _determine_csv_parameters(
            csv_content=content,
            user_params=user_params
        )

        # Extract parameters to be used
        delimiter = detection_params['delimiter']
        decimal_sep = detection_params['decimal_separator']
        has_header = detection_params['has_header']

        # Update report with used parameters
        report['detection_params'] = {
            'delimiter': delimiter,
            'decimal_separator': decimal_sep,
            'has_header': has_header
        }
        report['delimiter'] = delimiter
        report['decimal_separator'] = decimal_sep
        report['has_header'] = has_header

        # --- 4) Load with pandas.read_csv using the determined parameters ---
        read_csv_kwargs = {
            'sep': delimiter,
            'decimal': decimal_sep,
            'header': 0 if has_header else None,
            'na_filter': True,
            'na_values': ['NA', 'N/A', ''],
            'keep_default_na': True,
            'engine': 'python',  # Start with python engine for better error reporting/flexibility
            'skip_blank_lines': True,
            'quoting': csv.QUOTE_MINIMAL,  # Default pandas behavior, explicit here
        }

        # Add user-provided read_csv args (besides the three we handled)
        for k, v in user_params.items():
            if k not in ['delimiter', 'decimal_separator', 'has_header']:
                read_csv_kwargs[k] = v

        try:
            # Pass the original content to StringIO
            data = pd.read_csv(io.StringIO(content), **read_csv_kwargs)
        except Exception as e1:
            print(f"Warning: read_csv with engine='python' failed: {e1}")
            try:
                read_csv_kwargs['engine'] = 'c'
                data = pd.read_csv(io.StringIO(content), **read_csv_kwargs)
            except Exception as e2:
                msg = f"Could not parse CSV. Python engine error: {e1} | C engine error: {e2}"
                report['error'] = msg
                return None, report, None
        report['initial_shape'] = data.shape

        # ---> FIX: Ensure column names are strings <---
        data.columns = data.columns.astype(str)

        report['shape_after_all_na_col_removal'] = data.shape

        if data.empty:  # If all columns were NA or file was effectively empty after header
            report['warnings'].append("Data is empty after removing all-NA columns or due to empty content.")
            # Return empty DataFrame, report, and an empty Series for na_row_mask
            return pd.DataFrame(), report, pd.Series(dtype=bool)

        # --- 5) Handle type conversion based on data_type ---
        # Ensure categorical_info is reset for this run, it's part of the main report dict.
        report['categorical_info'] = {}
        _local_categorical_mappings = {} # Use a local temporary dict for populating

        if data_type == 'y':
            for col in data.columns:
                _original_col_series = data[col].copy() # Keep original for astype(str) and original NaN count
                _numeric_representation = pd.to_numeric(data[col], errors='coerce') # For NaN comparison and default conversion

                _original_is_object = pd.api.types.is_object_dtype(data[col].dtype)
                _original_is_numeric = pd.api.types.is_numeric_dtype(data[col].dtype)

                if categorical_mode == 'auto':
                    should_treat_as_categorical = False
                    if _original_is_object:
                        should_treat_as_categorical = True
                    elif not _original_is_numeric:  # Catches mixed types, booleans etc.
                        # Heuristic: if to_numeric creates more NaNs than original, or low cardinality
                        if _numeric_representation.isna().sum() > _original_col_series.isna().sum() or \
                           (_original_col_series.nunique() < len(_original_col_series) * 0.8 and _original_col_series.nunique() < 50):
                            should_treat_as_categorical = True

                    if should_treat_as_categorical:
                        # Warning for ambiguous numeric-like headers
                        # Check if column name is purely numeric or float-like (one dot)
                        if col.isdigit() or (col.count('.') == 1 and col.replace('.', '', 1).isdigit()):
                            report['warnings'].append(f"Column '{col}' detected as categorical but has a numeric header")

                        # Factorize using the original data treated as strings to ensure correct categories
                        _codes, _categories = pd.factorize(_original_col_series.astype(str))
                        data[col] = _codes
                        _local_categorical_mappings[col] = {'categories': _categories.tolist()}
                    else:
                        # Not deemed categorical by 'auto' logic, so make it numeric
                        data[col] = _numeric_representation

                elif categorical_mode == 'preserve':
                    # In 'preserve' mode, all columns are converted to numeric.
                    # String values will become NaN. Numerics stay. No factorization.
                    data[col] = _numeric_representation
                    # _local_categorical_mappings remains empty for 'preserve'

                elif categorical_mode == 'none':
                    # In 'none' mode, all columns are converted to numeric.
                    # String values will become NaN. No factorization.
                    data[col] = _numeric_representation
                    # _local_categorical_mappings remains empty for 'none'

            report['categorical_info'] = _local_categorical_mappings # Assign collected mappings to report

        elif data_type == 'x':
            # For X data, all columns are converted to numeric, coercing errors
            for col in data.columns:
                data[col] = pd.to_numeric(data[col], errors='coerce')
            # report['categorical_info'] remains empty for data_type 'x' as it was cleared/initialized above

        # --- 6) Identify rows with NA values (POST type conversion) ---
        # This mask reflects NAs *after* all above conversions.
        # This is the mask that should be returned as the third element for potential synchronization by the caller.
        na_mask_after_conversions = data.isna().any(axis=1)
        report['na_handling']['na_detected_in_rows'] = bool(na_mask_after_conversions.any())

        # --- Handle NA policy internally for load_csv ---
        # This affects the 'data' DataFrame that will be returned by this function.
        if report['na_handling']['na_detected_in_rows']: # Check if there are any NAs to handle
            if na_policy == 'abort':
                # Find first NA for error reporting
                first_na_row_label_in_current_data = data.index[na_mask_after_conversions][0]
                first_na_col_name = data.loc[first_na_row_label_in_current_data].isna().idxmax()
                error_msg = (f"NA values detected after processing and na_policy is 'abort'. "
                            f"First NA found in column '{first_na_col_name}' (row label: {first_na_row_label_in_current_data}) "
                            f"in file {path}.")
                report['error'] = error_msg
                report['na_handling']['na_detected'] = True
                # Return None for data, and the na_mask_after_conversions (though caller might not use if error)
                return None, report, na_mask_after_conversions

            elif na_policy == 'remove':
                # Update report fields about the rows that are about to be removed
                report['na_handling']['na_detected'] = True  # NAs were found and are being handled by removal
                report['na_handling']['nb_removed_rows'] = int(na_mask_after_conversions.sum())
                report['na_handling']['removed_rows_indices'] = data.index[na_mask_after_conversions].tolist()

                # Actually modify the 'data' DataFrame
                data = data[~na_mask_after_conversions].copy() # Use .copy() to avoid SettingWithCopyWarning

        # If na_policy == 'remove' but no NAs were detected, report fields remain at their initialized values (0, [], False)

        # --- 7) Final preparation of return values ---
        # 'final_shape' should reflect the shape of the data being returned.
        report['final_shape'] = data.shape
        report['final_column_names'] = data.columns.tolist()

        # Return the 'data' (possibly with rows removed by this function if na_policy='remove')
        # and 'na_mask_after_conversions' (which is the mask *before* this function's internal NA removal).
        # data_array = data.to_numpy().astype(np.float32)
        return data, report, na_mask_after_conversions

    except FileNotFoundError as e:
        report['error'] = str(e)
        return None, report, None
    except ValueError as e:
        report['error'] = f"ValueError during processing: {e}"
        return None, report, None
    except Exception as e:
        # Catch any other unexpected error during loading/processing
        import traceback
        report['error'] = f"Unexpected error in load_csv: {e}\n{traceback.format_exc()}"
        return None, report, None



"""
OperatorRegistry.py
A registry for operator controllers in the nirs4all pipeline.
"""

from typing import Type
from .controller import OperatorController

CONTROLLER_REGISTRY = []
def register_controller(operator_cls: Type[OperatorController]):
    """Decorator to register a controller class."""
    global CONTROLLER_REGISTRY
    if not issubclass(operator_cls, OperatorController):
        raise TypeError(f"Operator class {operator_cls.__name__} must inherit from OperatorController")

    # Check if controller is already registered (avoid duplicates)
    if any(c.__name__ == operator_cls.__name__ and c.__module__ == operator_cls.__module__
           for c in CONTROLLER_REGISTRY):
        # print(f"Controller {operator_cls.__name__} already registered, skipping...")
        return operator_cls

    # print(f"Registering controller: {operator_cls.__name__}")
    CONTROLLER_REGISTRY.append(operator_cls)
    CONTROLLER_REGISTRY.sort(key=lambda c: c.priority)
    # print(f"Registry now has {len(CONTROLLER_REGISTRY)} controllers: {[c.__name__ for c in CONTROLLER_REGISTRY]}")
    return operator_cls

def reset_registry():
    """Reset the controller registry."""
    global CONTROLLER_REGISTRY
    CONTROLLER_REGISTRY = []
    # print("Controller registry has been reset.")
